# -*- coding: utf-8 -*-
import numpy as np
import warnings
from scipy.interpolate import pade
from scipy.special import factorial

from nengo.utils.compat import is_integer

from nengolib.signal.system import LinearSystem, s

__all__ = [
    'Lowpass', 'Alpha', 'DoubleExp', 'Bandpass', 'Highpass',
    'pade_delay_error', 'PadeDelay', 'LegendreDelay']


def Lowpass(tau):
    """A first-order lowpass: ``1/(tau*s + 1)``.

    Parameters
    ----------
    tau : ``float``
        Time-constant of exponential decay.

    Returns
    -------
    sys : :class:`.LinearSystem`
        First-order lowpass.

    See Also
    --------
    :class:`nengo.Lowpass`
    :attr:`.s`

    Examples
    --------
    >>> from nengolib import Lowpass
    >>> import matplotlib.pyplot as plt
    >>> taus = np.linspace(.01, .05, 5)
    >>> for tau in taus:
    >>>     sys = Lowpass(tau)
    >>>     plt.plot(sys.ntrange(100), sys.impulse(100),
    >>>              label=r"$\\tau=%s$" % tau)
    >>> plt.xlabel("Time (s)")
    >>> plt.legend()
    >>> plt.show()
    """

    return 1 / (tau*s + 1)


def Alpha(tau):
    """A second-order lowpass: ``1/(tau*s + 1)**2``.

    Equivalent to convolving two identical lowpass synapses together.

    Parameters
    ----------
    tau : ``float``
        Time-constant of exponential decay.

    Returns
    -------
    sys : :class:`.LinearSystem`
        Second-order lowpass with identical time-constants.

    See Also
    --------
    :class:`nengo.Alpha`
    :func:`.Lowpass`
    :func:`.DoubleExp`

    Examples
    --------
    >>> from nengolib import Alpha
    >>> import matplotlib.pyplot as plt
    >>> taus = np.linspace(.01, .05, 5)
    >>> for tau in taus:
    >>>     sys = Alpha(tau)
    >>>     plt.plot(sys.ntrange(100), sys.impulse(100),
    >>>              label=r"$\\tau=%s$" % tau)
    >>> plt.xlabel("Time (s)")
    >>> plt.legend()
    >>> plt.show()
    """

    return DoubleExp(tau, tau)


def DoubleExp(tau1, tau2):
    """A second-order lowpass: ``1/((tau1*s + 1)*(tau2*s + 1))``.

    Equivalent to convolving two lowpass synapses together with potentially
    different time-constants, in either order.

    Parameters
    ----------
    tau1 : ``float``
        Time-constant of one exponential decay.
    tau2 : ``float``
        Time-constant of another exponential decay.

    Returns
    -------
    sys : :class:`.LinearSystem`
        Second-order lowpass with potentially different time-constants.

    See Also
    --------
    :func:`.Lowpass`
    :func:`.Alpha`

    Examples
    --------
    >>> from nengolib import DoubleExp
    >>> import matplotlib.pyplot as plt
    >>> tau1 = .03
    >>> taus = np.linspace(.01, .05, 5)
    >>> plt.title(r"$\\tau_1=%s$" % tau1)
    >>> for tau2 in taus:
    >>>     sys = DoubleExp(tau1, tau2)
    >>>     plt.plot(sys.ntrange(100), sys.impulse(100),
    >>>              label=r"$\\tau_2=%s$" % tau2)
    >>> plt.xlabel("Time (s)")
    >>> plt.legend()
    >>> plt.show()
    """

    return Lowpass(tau1) * Lowpass(tau2)


def Bandpass(freq, Q):
    """A second-order bandpass with given frequency and width.

    Parameters
    ----------
    freq : ``float``
        Frequency (in hertz) of the peak of the bandpass.
    Q : ``float``
        Inversely proportional to width of bandpass.

    Returns
    -------
    sys : :class:`.LinearSystem`
        Second-order lowpass with complex poles.

    See Also
    --------
    :func:`nengo.networks.Oscillator`

    Notes
    -----
    The state of this system is isomorphic to a decaying 2--dimensional
    oscillator with speed given by ``freq`` and decay given by ``Q``.

    References
    ----------
    .. [#] http://www.analog.com/library/analogDialogue/archives/43-09/EDCh%208%20filter.pdf

    Examples
    --------
    Bandpass filters centered around 20 Hz with varying bandwidths:

    >>> from nengolib.synapses import Bandpass
    >>> freq = 20
    >>> Qs = np.linspace(4, 40, 5)

    Evaluate each impulse (time-domain) response:

    >>> import matplotlib.pyplot as plt
    >>> plt.subplot(121)
    >>> for Q in Qs:
    >>>     sys = Bandpass(freq, Q)
    >>>     plt.plot(sys.ntrange(1000), sys.impulse(1000),
    >>>              label=r"$Q=%s$" % Q)
    >>> plt.xlabel("Time (s)")
    >>> plt.legend()

    Evaluate each frequency responses:

    >>> plt.subplot(122)
    >>> freqs = np.linspace(0, 40, 100)  # to evaluate
    >>> for Q in Qs:
    >>>     sys = Bandpass(freq, Q)
    >>>     plt.plot(freqs, np.abs(sys.evaluate(freqs)),
    >>>              label=r"$Q=%s$" % Q)
    >>> plt.xlabel("Frequency (Hz)")
    >>> plt.legend()
    >>> plt.show()

    Evaluate each state-space impulse (trajectory) after balancing:

    >>> from nengolib.signal import balance
    >>> for Q in Qs:
    >>>     plt.plot(*balance(Bandpass(freq, Q)).X.impulse(1000).T,
    >>>              label=r"$Q=%s$" % Q)
    >>> plt.legend()
    >>> plt.axis('equal')
    >>> plt.show()
    """  # noqa: E501

    w_0 = freq * (2*np.pi)
    return 1 / ((s/w_0)**2 + s/(w_0*Q) + 1)


def Highpass(tau, order=1):
    """A differentiated lowpass of given order: ``(tau*s/(tau*s + 1))**order``.

    Equivalent to differentiating the input, scaling by ``tau``,
    lowpass filtering with time-constant ``tau``, and finally repeating
    this ``order`` times. The lowpass filter is required to make this causal.

    Parameters
    ----------
    tau : ``float``
        Time-constant of the lowpass filter, and highpass gain.
    order : ``integer``, optional
        Dimension of the resulting linear system. Defaults to ``1``.

    Returns
    -------
    sys : :class:`.LinearSystem`
        Highpass filter with time-constant ``tau`` and dimension ``order``.

    See Also
    --------
    :func:`.Lowpass`
    :attr:`.s`

    Examples
    --------
    >>> from nengolib.synapses import Highpass

    Evaluate the highpass in the frequency domain with a time-constant of 10 ms
    and with a variety of orders:

    >>> tau = 1e-2
    >>> orders = list(range(1, 9))
    >>> freqs = np.linspace(0, 50, 100)  # to evaluate

    >>> import matplotlib.pyplot as plt
    >>> plt.title(r"$\\tau=%s$" % tau)
    >>> for order in orders:
    >>>     sys = Highpass(tau, order)
    >>>     assert len(sys) == order
    >>>     plt.plot(freqs, np.abs(sys.evaluate(freqs)),
    >>>              label=r"order=%s" % order)
    >>> plt.xlabel("Frequency (Hz)")
    >>> plt.legend()
    >>> plt.show()
    """

    if order < 1 or not is_integer(order):
        raise ValueError("order (%s) must be integer >= 1" % order)
    return (tau*s * Lowpass(tau))**order


def _passthrough_delay(m, c):
    """Analytically derived state-space when p = q = m.

    We use this because it is numerically stable for high m.
    """
    j = np.arange(1, m+1, dtype=np.float64)
    u = (m + j) * (m - j + 1) / (c * j)

    A = np.zeros((m, m))
    B = np.zeros((m, 1))
    C = np.zeros((1, m))
    D = np.zeros((1,))

    A[0, :] = B[0, 0] = -u[0]
    A[1:, :-1][np.diag_indices(m-1)] = u[1:]
    D[0] = (-1)**m
    C[0, np.arange(m) % 2 == 0] = 2*D[0]
    return LinearSystem((A, B, C, D), analog=True)


def _proper_delay(q, c):
    """Analytically derived state-space when p = q - 1.

    We use this because it is numerically stable for high q
    and doesn't have a passthrough.
    """
    j = np.arange(q, dtype=np.float64)
    u = (q + j) * (q - j) / (c * (j + 1))

    A = np.zeros((q, q))
    B = np.zeros((q, 1))
    C = np.zeros((1, q))
    D = np.zeros((1,))

    A[0, :] = -u[0]
    B[0, 0] = u[0]
    A[1:, :-1][np.diag_indices(q-1)] = u[1:]
    C[0, :] = (j + 1) / float(q) * (-1) ** (q - 1 - j)
    return LinearSystem((A, B, C, D), analog=True)


def _pade_delay(p, q, c):
    """Numerically evaluated state-space using Pade approximants.

    This may have numerical issues for large values of p or q.
    """
    i = np.arange(1, p+q+1, dtype=np.float64)
    taylor = np.append([1.0], (-c)**i / factorial(i))
    num, den = pade(taylor, q)
    return LinearSystem((num, den), analog=True)


def pade_delay_error(theta_times_freq, order, p=None):
    """Computes the approximation error in :func:`.PadeDelay`.

    For a given order, the difficulty of the delay is a function of the
    input frequency (:math:`s = 2j \\pi f`) times the delay length
    (:math:`\\theta`).

    Parameters
    ----------
    theta_times_freq : ``array_like``
        A float or array of floats (delay length times frequency) at which to
        evaluate the error.
    order : ``integer``
        ``order`` parameter passed to :func:`.PadeDelay`.
    p : ``integer``, optional
        ``p`` parameter passed to :func:`.PadeDelay`. Defaults to ``None``.

    Returns
    -------
    error : ``np.array`` of ``np.complex``
        Shaped like ``theta_times_freq``, with each element corresponding
        to the complex error term

        .. math::

            F(2j \\pi f) - e^{-\\theta \\times 2j \\pi f}

        where :math:`F(s)` is the transfer function constructed by
        :func:`.PadeDelay` for a delay of length :math:`\\theta`.

    See Also
    --------
    :func:`.PadeDelay`

    Examples
    --------
    >>> from nengolib.synapses import pade_delay_error
    >>> abs(pade_delay_error(1, order=6))
    0.0070350205992081461

    This means that for ``order=6`` and frequencies less than ``1/theta``,
    the approximation error is less than one percent!

    Now visualize the error across a range of frequencies, with various orders:

    >>> import matplotlib.pyplot as plt
    >>> freq_times_theta = np.linspace(0, 5, 1000)
    >>> for order in range(4, 9):
    >>>     plt.plot(freq_times_theta,
    >>>              abs(pade_delay_error(freq_times_theta, order=order)),
    >>>              label="order=%s" % order)
    >>> plt.xlabel(r"Frequency $\\times \\, \\theta$ (Unitless)")
    >>> plt.ylabel("Absolute Error")
    >>> plt.legend()
    >>> plt.show()
    """

    ttf = np.asarray(theta_times_freq)
    # switch to a delay of 1 for simplicity
    # this works due to the substitution of variables: theta*s <-> 1*s'
    sys = PadeDelay(1., order, p=p)
    return sys.evaluate(ttf) - np.exp(-2j*np.pi*ttf)


def _check_order(order):
    if order < 1 or not is_integer(order):
        raise ValueError("order (%s) must be integer >= 1" % order)
    return order


def PadeDelay(theta, order, p=None):
    """A finite-order approximation of a pure time-delay.

    Implements the transfer function:

    .. math::

       F(s) = e^{-\\theta s} + \\mathcal{O}(s^{\\texttt{order}+\\texttt{p}})

    or :math:`y(t) \\approx u(t - \\theta)` in the time-domain (for
    slowly changing inputs).

    This is the optimal approximation for a time-delay system given
    low-frequency inputs implemented using a finite-dimensional state.
    This is achieved via Padé approximants. The state-space of the
    system encodes a rolling window of input history
    (see :class:`.RollingWindow`). This can be used to approximate
    FIR filters and window functions in continuous time.

    Parameters
    ----------
    theta : ``float``
        Length of time-delay in seconds.
    order : ``integer``
        Order of approximation in the denominator
        (dimensionality of resulting system).
    p : ``integer``, optional
        Order of approximation in the numerator. Defaults to ``p=order-1``,
        since this gives the highest-order approximation without a passthrough.
        If ``p=order``, then the system will have a passthrough, which has
        a nonideal step response.

    Returns
    -------
    sys : :class:`.LinearSystem`
        Finite-order approximation of a pure time-delay.

    See Also
    --------
    :func:`.LegendreDelay`
    :func:`.pade_delay_error`
    :class:`.RollingWindow`
    :func:`.DiscreteDelay`
    :func:`scipy.interpolate.pade`

    Notes
    -----
    Closed-form derivations are found in [#]_.

    References
    ----------
    .. [#] A. R. Voelker and C. Eliasmith, "Improving spiking dynamical
       networks: Accurate delays, higher-order synapses, and time cells",
       2017, Submitted. [`URL <https://github.com/arvoelke/delay2017>`__]

    Examples
    --------
    >>> from nengolib.synapses import PadeDelay

    Delay 15 Hz band-limited white noise by 100 ms using various orders of
    approximations:

    >>> from nengolib.signal import z
    >>> from nengo.processes import WhiteSignal
    >>> import matplotlib.pyplot as plt
    >>> process = WhiteSignal(10., high=15, y0=0)
    >>> u = process.run_steps(500)
    >>> t = process.ntrange(len(u))
    >>> plt.plot(t, (z**-100).filt(u), linestyle='--', lw=4, label="Ideal")
    >>> for order in list(range(4, 9)):
    >>>     sys = PadeDelay(.1, order=order)
    >>>     assert len(sys) == order
    >>>     plt.plot(t, sys.filt(u), label="order=%s" % order)
    >>> plt.xlabel("Time (s)")
    >>> plt.legend()
    >>> plt.show()
    """

    q = _check_order(order)
    if p is None:
        p = q - 1

    if p < 1 or not is_integer(p):
        raise ValueError("p (%s) must be integer >= 1" % p)

    if p == q:
        return _passthrough_delay(p, theta)
    elif p == q - 1:
        return _proper_delay(q, theta)
    else:
        if q >= 10:
            warnings.warn("For large values of q (>= 10), p should either be "
                          "None, q - 1, or q.")
        return _pade_delay(p, q, theta)


def LegendreDelay(theta, order):
    """PadeDelay(theta, order) realizing the shifted Legendre basis.

    The transfer function is equivalent to :func:`.PadeDelay`, but its
    canonical state-space realization represents the window of history
    by the shifted Legendre polnomials:

    .. math::

       P_i(2 \\theta' \\theta^{-1} - 1)

    where ``i`` is the zero-based index into the state-vector.

    Parameters
    ----------
    theta : ``float``
        Length of time-delay in seconds.
    order : ``integer``
        Order of approximation in the denominator
        (dimensionality of resulting system).

    Returns
    -------
    sys : :class:`.LinearSystem`
        Finite-order approximation of a pure time-delay.

    See Also
    --------
    :func:`.PadeDelay`
    :func:`.pade_delay_error`
    :class:`.RollingWindow`

    Examples
    --------
    >>> from nengolib.synapses import LegendreDelay

    Delay 15 Hz band-limited white noise by 100 ms using various orders of
    approximations:

    >>> from nengolib.signal import z
    >>> from nengo.processes import WhiteSignal
    >>> import matplotlib.pyplot as plt
    >>> process = WhiteSignal(10., high=15, y0=0)
    >>> u = process.run_steps(500)
    >>> t = process.ntrange(len(u))
    >>> plt.plot(t, (z**-100).filt(u), linestyle='--', lw=4, label="Ideal")
    >>> for order in list(range(4, 9)):
    >>>     sys = LegendreDelay(.1, order=order)
    >>>     assert len(sys) == order
    >>>     plt.plot(t, sys.filt(u), label="order=%s" % order)
    >>> plt.xlabel("Time (s)")
    >>> plt.legend()
    >>> plt.show()
    """

    q = _check_order(order)

    Q = np.arange(q, dtype=np.float64)
    R = (2*Q + 1)[:, None] / theta
    j, i = np.meshgrid(Q, Q)

    A = np.where(i < j, -1, (-1.)**(i-j+1)) * R
    B = (-1.)**Q[:, None] * R
    C = np.ones((1, q))
    D = np.zeros((1,))

    return LinearSystem((A, B, C, D), analog=True)



from collections import defaultdict
import itertools

from .compat import iteritems
from .graphs import add_edges
from .stdlib import groupby


def operator_dependency_graph(operators):  # noqa: C901
    # -- all views of a base object in a particular dictionary
    by_base_sets = defaultdict(set)
    by_base_writes = defaultdict(set)
    by_base_reads = defaultdict(set)
    reads = defaultdict(list)
    sets = defaultdict(list)
    incs = defaultdict(list)
    ups = defaultdict(list)

    for op in operators:
        for sig in op.sets:
            by_base_sets[sig.base].add(sig)

        for sig in op.sets + op.incs:
            by_base_writes[sig.base].add(sig)

        for sig in op.reads:
            by_base_reads[sig.base].add(sig)

        for sig in op.reads:
            reads[sig].append(op)

        for sig in op.sets:
            sets[sig].append(op)

        for sig in op.incs:
            incs[sig].append(op)

        for sig in op.updates:
            ups[sig].append(op)

    validate_ops(sets, ups, incs)

    # -- Scheduling algorithm for serial evaluation:
    #    1) All sets on a given memory block
    #    2) All incs on a given memory block
    #    3) All reads on a given memory block
    #    4) All updates on a given memory block

    dg = {op: set() for op in operators}  # ops are nodes of the graph

    # -- incs depend on sets
    for sig, post_ops in iteritems(incs):
        pre_ops = list(sets[sig])
        for sig2 in by_base_sets[sig.base]:
            if sig.may_share_memory(sig2):
                pre_ops.extend(sets[sig2])
        add_edges(dg, itertools.product(set(pre_ops), post_ops))

    # -- reads depend on writes (sets and incs)
    for sig, post_ops in iteritems(reads):
        pre_ops = sets[sig] + incs[sig]
        for sig2 in by_base_writes[sig.base]:
            if sig.may_share_memory(sig2):
                pre_ops.extend(sets[sig2] + incs[sig2])
        add_edges(dg, itertools.product(set(pre_ops), post_ops))

    # -- updates depend on reads, sets, and incs.
    for sig, post_ops in iteritems(ups):
        pre_ops = sets[sig] + incs[sig] + reads[sig]
        for sig2 in by_base_reads[sig.base].union(by_base_writes[sig.base]):
            if sig.may_share_memory(sig2):
                pre_ops.extend(sets[sig2] + incs[sig2] + reads[sig2])
        add_edges(dg, itertools.product(set(pre_ops), post_ops))

    return dg


def validate_ops(sets, ups, incs):
    # -- assert that only one op sets any particular view
    for sig in sets:
        sig_sets = sets[sig] + (sets.get(sig.base, []) if sig.is_view else [])
        assert len(sig_sets) == 1, (sig, sig_sets)

    # -- assert that only one op updates any particular view
    for sig in ups:
        sig_ups = ups[sig] + (ups.get(sig.base, []) if sig.is_view else [])
        assert len(sig_ups) == 1, (sig, sig_ups)

    # -- assert that no two views are both set and aliased
    for _, base_group in groupby(sets, lambda x: x.base, hashable=True):
        for sig, sig2 in itertools.combinations(base_group, 2):
            assert not sig.may_share_memory(sig2), (
                "%s shares memory with %s" % (sig, sig2))

    # -- assert that no two views are both updated and aliased
    for _, base_group in groupby(ups, lambda x: x.base, hashable=True):
        for sig, sig2 in itertools.combinations(base_group, 2):
            assert not sig.may_share_memory(sig2), (
                "%s shares memory with %s" % (sig, sig2))



import os

import pytest
import _pytest.capture

from nengo.utils.paths import examples_dir
from nengo.utils.stdlib import execfile

# Monkeypatch _pytest.capture.DontReadFromInput
#  If we don't do this, importing IPython will choke as it reads the current
#  sys.stdin to figure out the encoding it will use; pytest installs
#  DontReadFromInput as sys.stdin to capture output.
#  Running with -s option doesn't have this issue, but this monkeypatch
#  doesn't have any side effects, so it's fine.
_pytest.capture.DontReadFromInput.encoding = "utf-8"
_pytest.capture.DontReadFromInput.write = lambda: None
_pytest.capture.DontReadFromInput.flush = lambda: None


too_slow = ['inhibitory_gating',
            'izhikevich',
            'learn_communication_channel',
            'learn_product',
            'learn_square',
            'learn_unsupervised',
            'lorenz_attractor',
            'nef_algorithm',
            'nef_summary',
            'network_design',
            'network_design_advanced',
            'question',
            'question_control',
            'question_memory',
            'spa_parser',
            'spa_sequence',
            'spa_sequence_routed']

all_examples, slow_examples, fast_examples = [], [], []

for subdir, _, files in os.walk(examples_dir):
    if (os.path.sep + '.') in subdir:
        continue
    files = [f for f in files if f.endswith('.ipynb')]
    examples = [os.path.join(subdir, os.path.splitext(f)[0]) for f in files]
    all_examples.extend(examples)
    slow_examples.extend([e for e, f in zip(examples, files)
                          if os.path.splitext(f)[0] in too_slow])
    fast_examples.extend([e for e, f in zip(examples, files)
                          if os.path.splitext(f)[0] not in too_slow])

# os.walk goes in arbitrary order, so sort after the fact to keep pytest happy
all_examples.sort()
slow_examples.sort()
fast_examples.sort()


def assert_noexceptions(nb_file, tmpdir):
    plt = pytest.importorskip('matplotlib.pyplot')
    pytest.importorskip("IPython", minversion="1.0")
    pytest.importorskip("jinja2")
    from nengo.utils.ipython import export_py, load_notebook
    nb_path = os.path.join(examples_dir, "%s.ipynb" % nb_file)
    nb = load_notebook(nb_path)
    pyfile = "%s.py" % (
        tmpdir.join(os.path.splitext(os.path.basename(nb_path))[0]))
    export_py(nb, pyfile)
    execfile(pyfile, {})
    plt.close('all')


@pytest.mark.example
@pytest.mark.parametrize('nb_file', fast_examples)
@pytest.mark.filterwarnings("ignore:Creating new attribute 'memory_location'")
def test_fast_noexceptions(nb_file, tmpdir):
    """Ensure that no cells raise an exception."""
    assert_noexceptions(nb_file, tmpdir)


@pytest.mark.slow
@pytest.mark.example
@pytest.mark.parametrize('nb_file', slow_examples)
def test_slow_noexceptions(nb_file, tmpdir):
    """Ensure that no cells raise an exception."""
    assert_noexceptions(nb_file, tmpdir)


def iter_cells(nb_file, cell_type="code"):
    from nengo.utils.ipython import load_notebook
    nb = load_notebook(os.path.join(examples_dir, "%s.ipynb" % nb_file))

    if nb.nbformat <= 3:
        cells = []
        for ws in nb.worksheets:
            cells.extend(ws.cells)
    else:
        cells = nb.cells

    for cell in cells:
        if cell.cell_type == cell_type:
            yield cell


@pytest.mark.example
@pytest.mark.parametrize('nb_file', all_examples)
def test_no_signature(nb_file):
    from nengo.utils.ipython import load_notebook
    nb = load_notebook(os.path.join(examples_dir, "%s.ipynb" % nb_file))
    assert 'signature' not in nb.metadata, "Notebook has signature"


@pytest.mark.example
@pytest.mark.parametrize('nb_file', all_examples)
def test_no_outputs(nb_file):
    """Ensure that no cells have output."""
    pytest.importorskip("IPython", minversion="1.0")

    for cell in iter_cells(nb_file):
        assert cell.outputs == [], "Cell outputs not cleared"



import collections

import numpy as np

from nengo.builder import Builder, Signal
from nengo.builder.ensemble import gen_eval_points, get_activities
from nengo.builder.node import SimPyFunc
from nengo.builder.operator import Copy, DotInc, ElementwiseInc, Reset
from nengo.connection import Connection
from nengo.dists import get_samples
from nengo.ensemble import Ensemble, Neurons
from nengo.exceptions import BuildError, ObsoleteError
from nengo.neurons import Direct
from nengo.node import Node
from nengo.solvers import NoSolver, Solver
from nengo.utils.compat import is_iterable, itervalues

built_attrs = ['eval_points', 'solver_info', 'weights', 'transform']


class BuiltConnection(collections.namedtuple('BuiltConnection', built_attrs)):
    """Collects the parameters generated in `.build_connection`.

    These are stored here because in the majority of cases the equivalent
    attribute in the original connection is a `.Distribution`. The attributes
    of a BuiltConnection are the full NumPy arrays used in the simulation.

    See the `.Connection` documentation for more details on each parameter.

    .. note:: The ``decoders`` attribute is obsolete as of Nengo 2.1.0.
              Use the ``weights`` attribute instead.

    Parameters
    ----------
    eval_points : ndarray
        Evaluation points.
    solver_info : dict
        Information dictionary returned by the `.Solver`.
    weights : ndarray
        Connection weights. May be synaptic connection weights defined in
        the connection's transform, or a combination of the decoders
        automatically solved for and the specified transform.
    transform : ndarray
        The transform matrix.
    """

    __slots__ = ()

    def __new__(cls, eval_points, solver_info, weights, transform):
        # Overridden to suppress the default __new__ docstring
        return tuple.__new__(
            cls, (eval_points, solver_info, weights, transform))

    @property
    def decoders(self):
        raise ObsoleteError("decoders are now part of 'weights'. "
                            "Access BuiltConnection.weights instead.",
                            since="v2.1.0")


def get_eval_points(model, conn, rng):
    if conn.eval_points is None:
        view = model.params[conn.pre_obj].eval_points.view()
        view.setflags(write=False)
        return view
    else:
        return gen_eval_points(
            conn.pre_obj, conn.eval_points, rng, conn.scale_eval_points)


def get_targets(conn, eval_points):
    if conn.function is None:
        targets = eval_points[:, conn.pre_slice]
    elif isinstance(conn.function, np.ndarray):
        targets = conn.function
    else:
        targets = np.zeros((len(eval_points), conn.size_mid))
        for i, ep in enumerate(eval_points[:, conn.pre_slice]):
            out = conn.function(ep)
            if out is None:
                raise BuildError("Building %s: Connection function returned "
                                 "None. Cannot solve for decoders." % (conn,))
            targets[i] = out

    return targets


def build_linear_system(model, conn, rng):
    eval_points = get_eval_points(model, conn, rng)
    ens = conn.pre_obj
    activities = get_activities(model.params[ens], ens, eval_points)
    if np.count_nonzero(activities) == 0:
        raise BuildError(
            "Building %s: 'activites' matrix is all zero for %s. "
            "This is because no evaluation points fall in the firing "
            "ranges of any neurons." % (conn, conn.pre_obj))

    targets = get_targets(conn, eval_points)
    return eval_points, activities, targets


def build_decoders(model, conn, rng, transform):
    encoders = model.params[conn.pre_obj].encoders
    gain = model.params[conn.pre_obj].gain
    bias = model.params[conn.pre_obj].bias

    eval_points = get_eval_points(model, conn, rng)
    targets = get_targets(conn, eval_points)

    x = np.dot(eval_points, encoders.T / conn.pre_obj.radius)
    E = None
    if conn.solver.weights:
        E = model.params[conn.post_obj].scaled_encoders.T[conn.post_slice]
        # include transform in solved weights
        targets = multiply(targets, transform.T)

    wrapped_solver = (model.decoder_cache.wrap_solver(solve_for_decoders)
                      if model.seeded[conn] else solve_for_decoders)
    decoders, solver_info = wrapped_solver(
        conn, gain, bias, x, targets, rng=rng, E=E)

    weights = (decoders.T if conn.solver.weights else
               multiply(transform, decoders.T))
    return eval_points, weights, solver_info


def solve_for_decoders(conn, gain, bias, x, targets, rng, E=None):
    activities = conn.pre_obj.neuron_type.rates(x, gain, bias)
    if np.count_nonzero(activities) == 0:
        raise BuildError(
            "Building %s: 'activities' matrix is all zero for %s. "
            "This is because no evaluation points fall in the firing "
            "ranges of any neurons." % (conn, conn.pre_obj))

    decoders, solver_info = conn.solver(activities, targets, rng=rng, E=E)
    return decoders, solver_info


def multiply(x, y):
    if x.ndim <= 2 and y.ndim < 2:
        return x * y
    elif x.ndim < 2 and y.ndim == 2:
        return x.reshape(-1, 1) * y
    elif x.ndim == 2 and y.ndim == 2:
        return np.dot(x, y)
    else:
        raise BuildError("Tensors not supported (x.ndim = %d, y.ndim = %d)"
                         % (x.ndim, y.ndim))


def slice_signal(model, signal, sl):
    assert signal.ndim == 1
    if isinstance(sl, slice) and (sl.step is None or sl.step == 1):
        return signal[sl]
    else:
        size = np.arange(signal.size)[sl].size
        sliced_signal = Signal(np.zeros(size), name="%s.sliced" % signal.name)
        model.add_op(Copy(signal, sliced_signal, src_slice=sl))
        return sliced_signal


@Builder.register(Solver)
def build_solver(model, solver, conn, rng, transform):
    return build_decoders(model, conn, rng, transform)


@Builder.register(NoSolver)
def build_no_solver(model, solver, conn, rng, transform):
    activities = np.zeros((1, conn.pre_obj.n_neurons))
    targets = np.zeros((1, conn.size_mid))
    E = np.zeros((1, conn.post_obj.n_neurons)) if solver.weights else None
    # No need to invoke the cache for NoSolver
    decoders, solver_info = conn.solver(activities, targets, rng=rng, E=E)
    weights = (decoders.T if conn.solver.weights else
               multiply(transform, decoders.T))
    return None, weights, solver_info


@Builder.register(Connection)  # noqa: C901
def build_connection(model, conn):
    """Builds a `.Connection` object into a model.

    A brief summary of what happens in the connection build process,
    in order:

    1. Solve for decoders.
    2. Combine transform matrix with decoders to get weights.
    3. Add operators for computing the function
       or multiplying neural activity by weights.
    4. Call build function for the synapse.
    5. Call build function for the learning rule.
    6. Add operator for applying learning rule delta to weights.

    Some of these steps may be altered or omitted depending on the parameters
    of the connection, in particular the pre and post types.

    Parameters
    ----------
    model : Model
        The model to build into.
    conn : Connection
        The connection to build.

    Notes
    -----
    Sets ``model.params[conn]`` to a `.BuiltConnection` instance.
    """

    # Create random number generator
    rng = np.random.RandomState(model.seeds[conn])

    # Get input and output connections from pre and post
    def get_prepost_signal(is_pre):
        target = conn.pre_obj if is_pre else conn.post_obj
        key = 'out' if is_pre else 'in'

        if target not in model.sig:
            raise BuildError("Building %s: the %r object %s is not in the "
                             "model, or has a size of zero."
                             % (conn, 'pre' if is_pre else 'post', target))
        if key not in model.sig[target]:
            raise BuildError(
                "Building %s: the %r object %s has a %r size of zero."
                % (conn, 'pre' if is_pre else 'post', target, key))

        return model.sig[target][key]

    model.sig[conn]['in'] = get_prepost_signal(is_pre=True)
    model.sig[conn]['out'] = get_prepost_signal(is_pre=False)

    weights = None
    eval_points = None
    solver_info = None
    signal_size = conn.size_out
    post_slice = conn.post_slice

    # Sample transform if given a distribution
    transform = get_samples(
        conn.transform, conn.size_out, d=conn.size_mid, rng=rng)

    # Figure out the signal going across this connection
    in_signal = model.sig[conn]['in']
    if (isinstance(conn.pre_obj, Node) or
            (isinstance(conn.pre_obj, Ensemble) and
             isinstance(conn.pre_obj.neuron_type, Direct))):
        # Node or Decoded connection in directmode
        weights = transform
        sliced_in = slice_signal(model, in_signal, conn.pre_slice)
        if conn.function is None:
            in_signal = sliced_in
        elif isinstance(conn.function, np.ndarray):
            raise BuildError("Cannot use function points in direct connection")
        else:
            in_signal = Signal(np.zeros(conn.size_mid), name='%s.func' % conn)
            model.add_op(SimPyFunc(in_signal, conn.function, None, sliced_in))
    elif isinstance(conn.pre_obj, Ensemble):  # Normal decoded connection
        eval_points, weights, solver_info = model.build(
            conn.solver, conn, rng, transform)
        if conn.solver.weights:
            model.sig[conn]['out'] = model.sig[conn.post_obj.neurons]['in']
            signal_size = conn.post_obj.neurons.size_in
            post_slice = None  # don't apply slice later
    else:
        weights = transform
        in_signal = slice_signal(model, in_signal, conn.pre_slice)

    # Add operator for applying weights
    model.sig[conn]['weights'] = Signal(
        weights, name="%s.weights" % conn, readonly=True)
    signal = Signal(np.zeros(signal_size), name="%s.weighted" % conn)
    model.add_op(Reset(signal))
    op = ElementwiseInc if weights.ndim < 2 else DotInc
    model.add_op(op(model.sig[conn]['weights'],
                    in_signal,
                    signal,
                    tag="%s.weights_elementwiseinc" % conn))

    # Add operator for filtering
    if conn.synapse is not None:
        signal = model.build(conn.synapse, signal)

    # Store the weighted-filtered output in case we want to probe it
    model.sig[conn]['weighted'] = signal

    if isinstance(conn.post_obj, Neurons):
        # Apply neuron gains (we don't need to do this if we're connecting to
        # an Ensemble, because the gains are rolled into the encoders)
        gains = Signal(model.params[conn.post_obj.ensemble].gain[post_slice],
                       name="%s.gains" % conn)
        model.add_op(ElementwiseInc(
            gains, signal, model.sig[conn]['out'][post_slice],
            tag="%s.gains_elementwiseinc" % conn))
    else:
        # Copy to the proper slice
        model.add_op(Copy(
            signal, model.sig[conn]['out'], dst_slice=post_slice,
            inc=True, tag="%s" % conn))

    # Build learning rules
    if conn.learning_rule is not None:
        rule = conn.learning_rule
        rule = [rule] if not is_iterable(rule) else rule
        targets = []
        for r in itervalues(rule) if isinstance(rule, dict) else rule:
            model.build(r)
            targets.append(r.modifies)

        if 'encoders' in targets:
            encoder_sig = model.sig[conn.post_obj]['encoders']
            encoder_sig.readonly = False
        if 'decoders' in targets or 'weights' in targets:
            if weights.ndim < 2:
                raise BuildError(
                    "'transform' must be a 2-dimensional array for learning")
            model.sig[conn]['weights'].readonly = False

    model.params[conn] = BuiltConnection(eval_points=eval_points,
                                         solver_info=solver_info,
                                         transform=transform,
                                         weights=weights)



import numpy as np
try:
    from scipy.linalg import solve_lyapunov as solve_continuous_lyapunov
except ImportError:  # pragma: no cover; github.com/scipy/scipy/pull/8082
    from scipy.linalg import solve_continuous_lyapunov
from scipy.linalg import solve_discrete_lyapunov, eig, cholesky, svd
from scipy.optimize import fminbound

from nengolib.signal.discrete import cont2discrete
from nengolib.signal.system import LinearSystem


__all__ = ['state_norm', 'control_gram', 'observe_gram',
           'balanced_transformation', 'hsvd', 'l1_norm']


def _H2P(A, B, analog):
    """Computes the positive-definite P matrix for determining the H2-norm."""
    if analog:
        P = solve_continuous_lyapunov(A, -np.dot(B, B.T))  # AP + PA^T = -BB^T
    else:
        # Note: discretization is not performed for the user
        P = solve_discrete_lyapunov(A, np.dot(B, B.T))  # APA^T - P = -BB^T
    return P


def state_norm(sys, norm='H2'):
    """Computes the norm of each dimension of ``x`` in the state-space.

    Parameters
    ----------
    sys : :data:`linear_system_like`
       Linear system representation.
    norm : ``string``, optional
       Defaults to ``'H2'``. Must be one of:

       * ``'H2'`` : The power of each dimension in response to white-noise
         input with uniform power, or equivalently in response to a delta
         impulse. [#]_

    Returns
    -------
    norms : ``(len(sys),) np.array``
       The norm of each dimension in ``sys``.

    See Also
    --------
    :class:`.H2Norm`
    :func:`.l1_norm`

    References
    ----------
    .. [#] http://www.maplesoft.com/support/help/maple/view.aspx?path=DynamicSystems%2FNormH2

    Examples
    --------
    >>> from nengolib.signal import state_norm
    >>> from nengolib.synapses import PadeDelay
    >>> sys = PadeDelay(.1, order=4)
    >>> dt = 1e-4
    >>> y = sys.X.impulse(2500, dt=dt)

    Comparing the analytical H2-norm of the delay state to its simulated value:

    >>> assert np.allclose(np.linalg.norm(y, axis=0) * np.sqrt(dt),
    >>>                    state_norm(sys), atol=1e-4)

    >>> import matplotlib.pyplot as plt
    >>> plt.plot(sys.ntrange(len(y), dt=dt), y)
    >>> plt.xlabel("Time (s)")
    >>> plt.show()
    """  # noqa: E501

    if norm == 'H2':
        # TODO: accept an additional sys describing the filtering on the input
        # so that we can get the norm in response to different input spectra.
        sys = LinearSystem(sys)
        A, B, C, D = sys.ss
        P = _H2P(A, B, sys.analog)
        return np.sqrt(P[np.diag_indices(len(P))])
    else:
        raise NotImplementedError("norm (%s) must be one of: 'H2'" % (norm,))


def control_gram(sys):
    """Computes the controllability/reachability gramiam of a linear system.

    Parameters
    ----------
    sys : :data:`linear_system_like`
       Linear system representation.

    Returns
    -------
    W : ``(len(sys), len(sys)) np.array``
       The controllability/reachability gramiam matrix.

    See Also
    --------
    :func:`.observe_gram`
    :func:`.balanced_transformation`

    References
    ----------
    .. [#] https://en.wikibooks.org/wiki/Control_Systems/Controllability_and_Observability
    """  # noqa: E501

    sys = LinearSystem(sys)
    A, B, C, D = sys.ss
    return _H2P(A, B, sys.analog)


def observe_gram(sys):
    """Computes the observability gramiam of a linear system.

    Parameters
    ----------
    sys : :data:`linear_system_like`
       Linear system representation.

    Returns
    -------
    W : ``(len(sys), len(sys)) np.array``
       The observability gramiam matrix.

    See Also
    --------
    :func:`.control_gram`
    :func:`.balanced_transformation`

    References
    ----------
    .. [#] https://en.wikibooks.org/wiki/Control_Systems/Controllability_and_Observability
    """  # noqa: E501

    sys = LinearSystem(sys)
    A, B, C, D = sys.ss
    return _H2P(A.T, C.T, sys.analog)


def balanced_transformation(sys):
    """Computes the balancing transformation, its inverse, and eigenvalues.

    Parameters
    ----------
    sys : :data:`linear_system_like`
       Linear system representation.

    Returns
    -------
    T : ``(len(sys), len(sys)) np.array``
       Similarity transformation matrix.
    Tinv : ``(len(sys), len(sys)) np.array``
       Inverse of similarity transformation matrix.
    s : ``(len(sys),) np.array``
       Hankel singular values (see :func:`.hsvd`).

    See Also
    --------
    :func:`.balance`
    :class:`.Balanced`
    :func:`.hsvd`
    :func:`.LinearSystem.transform`

    References
    ----------
    .. [#] Laub, A.J., M.T. Heath, C.C. Paige, and R.C. Ward, "Computation of
       System Balancing Transformations and Other Applications of
       Simultaneous Diagonalization Algorithms," *IEEE Trans. Automatic
       Control*, AC-32 (1987), pp. 115-122.

    .. [#] http://www.mathworks.com/help/control/ref/balreal.html
    """

    sys = LinearSystem(sys)  # cast first to memoize sys2ss in control_gram
    if not sys.analog:
        raise NotImplementedError("balanced digital filters not supported")

    R = control_gram(sys)
    O = observe_gram(sys)

    LR = cholesky(R, lower=True)
    LO = cholesky(O, lower=True)

    U, S, V = svd(np.dot(LO.T, LR))

    T = np.dot(LR, V.T) * S ** (-1. / 2)
    Tinv = (S ** (-1. / 2))[:, None] * np.dot(U.T, LO.T)

    return T, Tinv, S


def hsvd(sys):
    """Compute Hankel singular values of a linear system.

    Parameters
    ----------
    sys : :data:`linear_system_like`
       Linear system representation.

    Returns
    -------
    s : ``(len(sys),) np.array``
       Hankel singular values.

    See Also
    --------
    :class:`.Hankel`
    :func:`.balanced_transformation`

    References
    ----------
    .. [#] Glover, Keith, and Jonathan R. Partington. "Bounds on the
       achievable accuracy in model reduction." Modelling, robustness and
       sensitivity reduction in control systems. Springer Berlin
       Heidelberg, 1987. 95-118.

    .. [#] https://www.mathworks.com/help/control/ref/hsvd.html
    """

    sys = LinearSystem(sys)
    R, O = control_gram(sys), observe_gram(sys)
    # sort needed to be consistent across different versions
    return np.sort(np.sqrt(abs(eig(np.dot(O, R))[0])))[::-1]


def _state_impulse(A, x0, k, delay=0):
    """Computes the states for discrete feedback A starting from x0."""
    x = np.zeros((k, len(A)))
    x[delay, :] = np.squeeze(x0)
    for i in range(delay+1, k):
        x[i, :] = np.dot(A, x[i-1, :])  # Note: doesn't assume canonical form
    return x


def l1_norm(sys, rtol=1e-6, max_length=2**18):
    """Computes the L1-norm of a linear system within a relative tolerance.

    The L1-norm of a (BIBO stable) linear system is the integral of the
    absolute value of its impulse response. For unstable systems this will be
    infinite. The L1-norm is important because it bounds the worst-case
    output of the system for arbitrary inputs within ``[-1, 1]``. In fact,
    this worst-case output is achieved by reversing the input which alternates
    between ``-1`` and ``1`` during the intervals where the impulse response is
    negative or positive, respectively (in the limit as ``T -> infinity``).

    Algorithm adapted from [#]_ following the methods of [#]_. This works by
    iteratively refining lower and upper bounds using progressively longer
    simulations and smaller timesteps. The lower bound is given by the
    absolute values of the discretized response. The upper bound is given by
    refining the time-step intervals where zero-crossings may have occurred.

    Parameters
    ----------
    sys : :data:`linear_system_like`
       Linear system representation.
    rtol : ``float``, optional
       Desired error (relative tolerance).
       Smaller tolerances require more compute time.
       Defaults to ``1e-6``.
    max_length : ``integer``, optional
       Maximum number of time-steps to simulate the system's impulse response.
       The simulation time-step is varied by the algorithm.
       Defaults to ``2**18``.

    Returns
    -------
    norm : ``float``
       L1-norm of the output.

    See Also
    --------
    :class:`.L1Norm`
    :func:`.state_norm`

    Notes
    -----
    The algorithm will terminate after either ``rtol`` tolerance is met, or
    ``max_length`` simulation steps are required --- whichever occurs first.

    References
    ----------
    .. [#] http://www.mathworks.com/matlabcentral/fileexchange/41587-system-l1-norm/content/l1norm.m
       J.F. Whidborne (April 28, 1995).

    .. [#] Rutland, Neil K., and Paul G. Lane. "Computing the 1-norm of the
       impulse response of linear time-invariant systems."
       Systems & control letters 26.3 (1995): 211-221.
    """  # noqa: E501

    sys = LinearSystem(sys)
    if not sys.analog:
        raise ValueError("system (%s) must be analog" % sys)

    # Setup state-space system and check stability/conditioning
    # we will subtract out D and add it back in at the end
    A, B, C, D = sys.ss
    alpha = np.max(eig(A)[0].real)  # eq (28)
    if alpha >= 0:
        raise ValueError("system (%s) has unstable eigenvalue: %s" % (
            sys, alpha))

    # Compute a suitable lower-bound for the L1-norm
    # using the steady state response, which is equivalent to the
    # L1-norm without an absolute value (i.e. just an integral)
    G0 = sys.dcgain - sys.D  # -C.inv(A).B

    # Compute a suitable upper-bound for the L1-norm
    # Note this should be tighter than 2*sum(abs(hankel(sys)))
    def _normtail(sig, A, x, C):
        # observability gramiam when A perturbed by sig
        W = solve_continuous_lyapunov(A.T + sig*np.eye(len(A)), -C.T.dot(C))
        return np.sqrt(x.dot(W).dot(x.T) / 2 / sig)  # eq (39)

    xtol = -alpha * 1e-4
    _, fopt, _, _ = fminbound(
        _normtail, 0, -alpha, (A, B.T, C), xtol=xtol, full_output=True)

    # Setup parameters for iterative optimization
    L, U = abs(G0), fopt
    N = 2**4
    T = -1 / alpha

    while N <= max_length and .5 * (U - L) / L >= rtol:  # eq (25)

        # Step 1. Improve the lower bound by simulating more.
        dt = T / N
        dsys = cont2discrete((A, B, C, 0), dt=dt)
        Phi = dsys.A

        y = dsys.impulse(N)
        abs_y = abs(y)
        L_impulse = np.sum(abs_y)

        # bound the missing response from t > T from below
        L_tail = abs(G0 - np.sum(y))  # eq (33)
        L = max(L, L_impulse + L_tail)

        # Step 2. Improve the upper bound using refined interval method.
        x = _state_impulse(Phi, x0=B, k=N, delay=0)  # eq (38)
        abs_e = np.squeeze(abs(C.dot(x.T)))
        x = x[:-1]  # compensate for computing where thresholds crossed
        abs_y = abs_y[:-1]  # compensate for computing where thresholds crossed

        # find intervals that could have zero-crossings and adjust their
        # upper bounds (the lower bound is exact for the other intervals)
        CTC = C.T.dot(C)
        W = solve_continuous_lyapunov(
            A.T, Phi.T.dot(CTC).dot(Phi) - CTC)  # eq (36)
        AWA = A.T.dot(W.dot(A))
        thresh = np.squeeze(  # eq (41)
            np.sqrt(dt * np.sum(x.dot(AWA) * x, axis=1)))
        cross = np.maximum(abs_e[:-1], abs_e[1:]) <= thresh  # eq (20)
        abs_y[cross] = np.sqrt(  # eq (22, 37)
            dt * np.sum(x[cross].dot(W) * x[cross], axis=1))

        # bound the missing response from t > T from above
        _, U_tail, _, _ = fminbound(
            _normtail, 0, -alpha, (A, x[-1], C), xtol=xtol, full_output=True)
        U_impulse = np.sum(abs_y)
        U = max(min(U, U_impulse + U_tail), L)

        N *= 2
        if U_impulse - L_impulse < U_tail - L_tail:  # eq (26)
            T *= 2

    return (U + L) / 2 + abs(np.asarray(D).squeeze()), .5 * (U - L) / L



"""
Signal (:mod:`nengolib.signal`)
===============================

.. currentmodule:: nengolib.signal

Linear Systems
--------------

.. autosummary::
   :toctree:

   LinearSystem
   cont2discrete
   discrete2cont

Model Reduction
---------------

.. autosummary::
   :toctree:

   pole_zero_cancel
   modred
   balance
   balred

Realizations
------------

.. autosummary::
   :toctree:

   Identity
   Balanced
   Hankel
   L1Norm
   H2Norm

Distributions
-------------

.. autosummary::
   :toctree:

   EvalPoints
   Encoders

Learning
--------

.. autosummary::
   :toctree:

   pes_learning_rate

Lyapunov Theory
---------------

.. autosummary::
   :toctree:

   l1_norm
   state_norm
   control_gram
   observe_gram
   hsvd
   balanced_transformation

Miscellaneous
-------------

.. autosummary::
   :toctree:

   nrmse
   shift
"""

from .discrete import *
from .dists import *
from .learning import *
from .lyapunov import *
from .realizers import *
from .reduction import *
from .system import *
from .utils import *



import pytest

import nengo
import nengo.synapses
from nengo.exceptions import ConfigError, ReadonlyError
from nengo.params import Parameter
from nengo.utils.testing import ThreadedAssertion


def test_config_basic():
    model = nengo.Network()
    model.config[nengo.Ensemble].set_param('something',
                                           Parameter('something', None))
    model.config[nengo.Ensemble].set_param('other',
                                           Parameter('other', default=0))
    model.config[nengo.Connection].set_param('something_else',
                                             Parameter('something_else', None))

    with pytest.raises(ConfigError):
        model.config[nengo.Ensemble].set_param('fails', 1.0)

    with model:
        a = nengo.Ensemble(50, dimensions=1)
        b = nengo.Ensemble(90, dimensions=1)
        a2b = nengo.Connection(a, b, synapse=0.01)

    with pytest.raises(ConfigError):
        model.config[a].set_param('thing', Parameter('thing', None))

    assert model.config[a].something is None
    assert model.config[b].something is None
    assert model.config[a].other == 0
    assert model.config[b].other == 0
    assert model.config[a2b].something_else is None

    model.config[a].something = 'hello'
    assert model.config[a].something == 'hello'
    model.config[a].something = 'world'
    assert model.config[a].something == 'world'
    del model.config[a].something
    assert model.config[a].something is None

    with pytest.raises(AttributeError):
        model.config[a].something_else
        model.config[a2b].something
    with pytest.raises(AttributeError):
        model.config[a].something_else = 1
        model.config[a2b].something = 1

    with pytest.raises(ConfigError):
        model.config['a'].something
    with pytest.raises(ConfigError):
        model.config[None].something


def test_network_nesting():
    """Make sure nested networks inherit configs."""
    with nengo.Network() as net1:
        # We'll change radius and seed. Make sure we have the right defaults.
        assert nengo.Ensemble.radius.default == 1.0
        assert nengo.Ensemble.seed.default is None

        # Before = default; after = what we set
        assert net1.config[nengo.Ensemble].radius == 1.0
        assert net1.config[nengo.Ensemble].seed is None
        net1.config[nengo.Ensemble].radius = 3.0
        net1.config[nengo.Ensemble].seed = 10
        assert net1.config[nengo.Ensemble].radius == 3.0
        assert net1.config[nengo.Ensemble].seed == 10

        # If we make an ensemble, it uses what we set, of the config default
        ens1 = nengo.Ensemble(10, 1, radius=2.0)
        assert ens1.seed == 10
        assert ens1.radius == 2.0

        # It's an error to configure an actual param with the config
        with pytest.raises(ConfigError):
            net1.config[ens1].radius = 3.0

        with nengo.Network() as net2:
            # We'll just change radius in net2.

            # Before = default; after = what we set
            assert net2.config[nengo.Ensemble].radius == 1.0
            net2.config[nengo.Ensemble].radius = 5.0
            assert net2.config[nengo.Ensemble].radius == 5.0

            # If we make an ensemble, it traverses the context stack
            ens2 = nengo.Ensemble(10, 1)
            assert ens2.radius == 5.0
            assert ens2.seed == 10

            with nengo.Network() as net3:
                # Works for > 1 levels
                net3.config[nengo.Ensemble].seed = 20
                ens3 = nengo.Ensemble(10, 1)
                assert ens3.seed == 20
                assert ens3.radius == 5.0


def test_context_is_threadsafe():
    class CheckIndependence(ThreadedAssertion):
        def init_thread(self, worker):
            setattr(worker, 'model', nengo.Network())
            worker.model.__enter__()

        def assert_thread(self, worker):
            assert list(nengo.Config.context) == [worker.model.config]

        def finish_thread(self, worker):
            worker.model.__exit__(*worker.exc_info)

    CheckIndependence(n_threads=2)


def test_defaults():
    """Test that settings defaults propagates appropriately."""
    b = nengo.Ensemble(10, dimensions=1, radius=nengo.Default,
                       add_to_container=False)

    assert b.radius == nengo.Ensemble.radius.default

    with nengo.Network():
        c = nengo.Ensemble(10, dimensions=1, radius=nengo.Default)
        with nengo.Network() as net2:
            net2.config[nengo.Ensemble].radius = 2.0
            a = nengo.Ensemble(50, dimensions=1, radius=nengo.Default)
            del net2.config[nengo.Ensemble].radius
            d = nengo.Ensemble(50, dimensions=1, radius=nengo.Default)

    assert c.radius == nengo.Ensemble.radius.default
    assert a.radius == 2.0
    assert d.radius == nengo.Ensemble.radius.default


def test_configstack():
    """Test that setting defaults with bare configs works."""
    with nengo.Network() as net:
        net.config[nengo.Connection].transform = -1
        e1 = nengo.Ensemble(5, dimensions=1)
        e2 = nengo.Ensemble(6, dimensions=1)
        excite = nengo.Connection(e1, e2)
        with nengo.Config(nengo.Connection) as inhib:
            inhib[nengo.Connection].synapse = nengo.synapses.Lowpass(0.00848)
            inhibit = nengo.Connection(e1, e2)
    assert excite.synapse == nengo.Connection.synapse.default
    assert excite.transform == -1
    assert inhibit.synapse == inhib[nengo.Connection].synapse
    assert inhibit.transform == -1


def test_config_property():
    """Test that config can't be easily modified."""
    with nengo.Network() as net:
        with pytest.raises(ReadonlyError):
            net.config = nengo.config.Config()
        with pytest.raises(AttributeError):
            del net.config
        assert nengo.config.Config.context[-1] is net.config
    assert len(nengo.config.Config.context) == 0


def test_external_class():
    class A(object):
        thing = Parameter('thing', default='hey')

    inst = A()
    config = nengo.Config(A)
    config[A].set_param('amount', Parameter('amount', default=1))

    # Extra param
    assert config[inst].amount == 1

    # Default still works like Nengo object
    assert inst.thing == 'hey'
    with pytest.raises(ConfigError):
        config[inst].thing


def test_instance_fallthrough():
    """If the class default is set, instances should use that."""
    class A(object):
        pass

    inst1 = A()
    inst2 = A()
    config = nengo.Config(A)
    config[A].set_param('amount', Parameter('amount', default=1))
    assert config[A].amount == 1
    assert config[inst1].amount == 1
    assert config[inst2].amount == 1
    # Value can change for instance
    config[inst1].amount = 2
    assert config[A].amount == 1
    assert config[inst1].amount == 2
    assert config[inst2].amount == 1
    # If value to A is changed, unset instances should also change
    config[A].amount = 3
    assert config[A].amount == 3
    assert config[inst1].amount == 2
    assert config[inst2].amount == 3
    # If class default is deleted, unset instances go back
    del config[A].amount
    assert config[A].amount == 1
    assert config[inst1].amount == 2
    assert config[inst2].amount == 1


def test_contains():
    class A(object):
        pass

    cfg = nengo.Config(A)
    with pytest.raises(TypeError):
        A in cfg



import warnings

import numpy as np

import nengo
from nengo.dists import Choice, Exponential, Uniform
from nengo.exceptions import ValidationError
from nengo.utils.compat import is_iterable, range
from nengo.utils.network import with_self
from .ensemblearray import EnsembleArray


def filtered_step(t, shift=0.5, scale=50, step_val=1):
    return np.maximum(-1 / np.exp((t - shift) * scale) + 1, 0) * step_val


class AssociativeMemory(nengo.Network):
    """Associative memory network.

    Parameters
    ----------
    input_vectors: array_like
        The list of vectors to be compared against.
    output_vectors: array_like, optional (Default: None)
        The list of vectors to be produced for each match. If None, the
        associative memory will be autoassociative (cleanup memory).
    n_neurons: int, optional (Default: 50)
        The number of neurons for each of the ensemble (where each ensemble
        represents each item in the input_vectors list).
    threshold: float, optional (Default: 0.3)
        The association activation threshold.
    input_scales: float or array_like, optional (Default: 1.0)
        Scaling factor to apply on each of the input vectors. Note that it
        is possible to scale each vector independently.
    inhibitable: bool, optional (Default: False)
        Flag to indicate if the entire associative memory module is
        inhibitable (entire thing can be shut off). The input gain into
        the inhibitory connection is 1.5.
    label : str, optional (Default: None)
        A name for the ensemble. Used for debugging and visualization.
    seed : int, optional (Default: None)
        The seed used for random number generation.
    add_to_container : bool, optional (Default: None)
        Determines if the network will be added to the current container.
        If None, will be true if currently within a Network.
    """

    exp_scale = 0.15  # Scaling factor for exponential distribution
    n_eval_points = 5000

    def __init__(self, input_vectors, output_vectors=None,  # noqa: C901
                 n_neurons=50, threshold=0.3, input_scales=1.0,
                 inhibitable=False,
                 label=None, seed=None, add_to_container=None):
        super(AssociativeMemory, self).__init__(label, seed, add_to_container)

        # --- Put arguments in canonical form
        if output_vectors is None:
            # If output vocabulary is not specified, use input vector list
            # (i.e autoassociative memory)
            output_vectors = input_vectors
        if is_iterable(input_vectors):
            input_vectors = np.array(input_vectors, ndmin=2)
        if is_iterable(output_vectors):
            output_vectors = np.array(output_vectors, ndmin=2)

        if input_vectors.shape[0] == 0:
            raise ValidationError("Number of input vectors cannot be 0.",
                                  attr='input_vectors', obj=self)
        elif input_vectors.shape[0] != output_vectors.shape[0]:
            # Fail if number of input items and number of output items don't
            # match
            raise ValidationError(
                "Number of input vectors does not match number of output "
                "vectors. %d != %d"
                % (input_vectors.shape[0], output_vectors.shape[0]),
                attr='input_vectors', obj=type(self))

        # Handle possible different threshold / input_scale values for each
        # element in the associative memory
        if not is_iterable(threshold):
            threshold = threshold * np.ones(input_vectors.shape[0])
        else:
            threshold = np.array(threshold)

        # --- Check preconditions
        self.n_items = input_vectors.shape[0]
        if self.n_items != output_vectors.shape[0]:
            raise ValidationError(
                "Number of input vectors (%d) does not match number of output "
                "vectors (%d)" % (self.n_items, output_vectors.shape[0]),
                attr='input_vectors', obj=self)
        if threshold.shape[0] != self.n_items:
            raise ValidationError(
                "Number of threshold values (%d) does not match number of "
                "input vectors (%d)." % (threshold.shape[0], self.n_items),
                attr='threshold', obj=self)

        # --- Set parameters
        self.out_conns = []  # Used in `add_threshold_to_output`
        # Used in `add_threshold_to_output`
        self.default_vector_inhibit_conns = []
        self.thresh_ens = None  # Will hold thresholded outputs
        self.is_wta = False
        self._inhib_scale = 1.5

        # -- Create the core network
        with self, self.am_ens_config:
            self.bias_node = nengo.Node(output=1)
            self.elem_input = nengo.Node(
                size_in=self.n_items, label="element input")
            self.elem_output = nengo.Node(
                size_in=self.n_items, label="element output")
            self.utilities = self.elem_output

            self.am_ensembles = []
            label_prefix = "" if label is None else label + "_"
            filt_scale = 15
            filt_step_func = lambda x: filtered_step(x, 0.0, scale=filt_scale)
            for i in range(self.n_items):
                e = nengo.Ensemble(n_neurons, 1, label=label_prefix + str(i))
                self.am_ensembles.append(e)

                # Connect input and output nodes
                nengo.Connection(self.bias_node, e, transform=-threshold[i])
                nengo.Connection(self.elem_input[i], e)
                nengo.Connection(
                    e, self.elem_output[i], function=filt_step_func)

            if inhibitable:
                # Input node for inhibitory gating signal (if enabled)
                self.inhibit = nengo.Node(size_in=1, label="inhibit")
                nengo.Connection(self.inhibit, self.elem_input,
                                 transform=-np.ones((self.n_items, 1))
                                 * self._inhib_scale)
                # Note: We can use a decoded connection here because all the
                # am_ensembles have [1] encoders
            else:
                self.inhibit = None
        self.add_input_mapping("input", input_vectors, input_scales)
        self.add_output_mapping("output", output_vectors)

    @property
    def am_ens_config(self):
        """(Config) Defaults for associative memory ensemble creation."""
        cfg = nengo.Config(nengo.Ensemble, nengo.Connection)
        cfg[nengo.Ensemble].update({
            'radius': 1,
            'intercepts': Exponential(self.exp_scale, 0.0, 1.0),
            'encoders': Choice([[1]]),
            'eval_points': Uniform(0.0, 1.0),
            'n_eval_points': self.n_eval_points,
        })
        cfg[nengo.Connection].synapse = None
        return cfg

    @property
    def default_ens_config(self):
        """(Config) Defaults for other ensemble creation."""
        cfg = nengo.Config(nengo.Ensemble)
        cfg[nengo.Ensemble].update({
            'radius': 1,
            'intercepts': Exponential(self.exp_scale, 0.0, 1.0),
            'encoders': Choice([[1]]),
            'eval_points': Uniform(0.0, 1.0),
            'n_eval_points': self.n_eval_points,
        })
        return cfg

    @property
    def thresh_ens_config(self):
        """(Config) Defaults for threshold ensemble creation."""
        cfg = nengo.Config(nengo.Ensemble)
        cfg[nengo.Ensemble].update({
            'radius': 1,
            'intercepts': Uniform(0.5, 1.0),
            'encoders': Choice([[1]]),
            'eval_points': Uniform(0.75, 1.1),
            'n_eval_points': self.n_eval_points,
        })
        return cfg

    @with_self
    def add_input_mapping(self, name, input_vectors, input_scales=1.0):
        """Adds a set of input vectors to the associative memory network.

        Creates a transform with the given input vectors between the
        a named input node and associative memory element input to enable the
        inputs to be mapped onto ensembles of the Associative Memory.

        Parameters
        ----------
        name: str
            Name to use for the input node. This name will be used as the name
            of the attribute for the associative memory network.
        input_vectors: array_like
            The list of vectors to be compared against.
        input_scales: float or array_like, optional (Default: 1.0)
            Scaling factor to apply on each of the input vectors. Note that it
            is possible to scale each vector independently.
        """
        # --- Put arguments in canonical form
        n_vectors, d_vectors = input_vectors.shape
        if is_iterable(input_vectors):
            input_vectors = np.array(input_vectors, ndmin=2)
        if not is_iterable(input_scales):
            input_scales = input_scales * np.ones((1, n_vectors))
        else:
            input_scales = np.array(input_scales, ndmin=2)

        # --- Check some preconditions
        if input_scales.shape[1] != n_vectors:
            raise ValidationError("Number of input_scale values (%d) does not "
                                  "match number of input vectors (%d)."
                                  % (input_scales.shape[1], n_vectors),
                                  attr='input_scales')
        if hasattr(self, name):
            raise ValidationError("Name '%s' already exists as a node in the "
                                  "associative memory." % name, attr='name')

        # --- Finally, make the input node and connect it
        in_node = nengo.Node(size_in=d_vectors, label=name)
        setattr(self, name, in_node)
        nengo.Connection(in_node, self.elem_input,
                         synapse=None,
                         transform=input_vectors * input_scales.T)

    @with_self
    def add_output_mapping(self, name, output_vectors):
        """Adds another output to the associative memory network.

        Creates a transform with the given output vectors between the
        associative memory element output and a named output node to enable the
        selection of output vectors by the associative memory.

        Parameters
        ----------
        name: str
            Name to use for the output node. This name will be used as
            the name of the attribute for the associative memory network.
        output_vectors: array_like
            The list of vectors to be produced for each match.
        """
        # --- Put arguments in canonical form
        if is_iterable(output_vectors):
            output_vectors = np.array(output_vectors, ndmin=2)

        # --- Check preconditions
        if hasattr(self, name):
            raise ValidationError("Name '%s' already exists as a node in the "
                                  "associative memory." % name, attr='name')

        # --- Make the output node and connect it
        output = nengo.Node(size_in=output_vectors.shape[1], label=name)
        setattr(self, name, output)

        if self.thresh_ens is not None:
            c = nengo.Connection(self.thresh_ens.output, output,
                                 synapse=None, transform=output_vectors.T)
        else:
            c = nengo.Connection(self.elem_output, output,
                                 synapse=None, transform=output_vectors.T)
        self.out_conns.append(c)

    @with_self
    def add_default_output_vector(self, output_vector, output_name='output',
                                  n_neurons=50, min_activation_value=0.5):
        """Adds a default output vector to the associative memory network.

        The default output vector is chosen if the input matches none of
        the given input vectors.

        Parameters
        ----------
        output_vector: array_like
            The vector to be produced if the input value matches none of
            the vectors in the input vector list.
        output_name: str, optional (Default: 'output')
            The name of the input to which the default output vector
            should be applied.
        n_neurons: int, optional (Default: 50)
            Number of neurons to use for the default output vector ensemble.
        min_activation_value: float, optional (Default: 0.5)
            Minimum activation value (i.e. threshold) to use to disable
            the default output vector.
        """
        with self.default_ens_config:
            default_vector_ens = nengo.Ensemble(
                n_neurons, 1, label="Default %s vector" % output_name)
            nengo.Connection(self.bias_node, default_vector_ens, synapse=None)

            tr = -(1.0 / min_activation_value) * np.ones((1, self.n_items))
            if self.thresh_ens is not None:
                c = nengo.Connection(
                    self.thresh_ens.output, default_vector_ens, transform=tr)
            else:
                c = nengo.Connection(
                    self.elem_output, default_vector_ens, transform=tr)

            # Add the output connection to the output connection list
            self.default_vector_inhibit_conns.append(c)

            # Make new output class attribute and connect to it
            output = getattr(self, output_name)
            nengo.Connection(default_vector_ens, output,
                             transform=np.array(output_vector, ndmin=2).T,
                             synapse=None)

            if self.inhibit is not None:
                nengo.Connection(self.inhibit, default_vector_ens,
                                 transform=-self._inhib_scale, synapse=None)

    @with_self
    def add_wta_network(self, inhibit_scale=1.5, inhibit_synapse=0.005):
        """Add a winner-take-all (WTA) network to associative memory output.

        Parameters
        ----------
        inhibit_scale: float, optional (Default: 1.5)
            Mutual inhibition scaling factor.
        inhibit_synapse: float, optional (Default: 0.005)
            Mutual inhibition synapse time constant.
        """
        if not self.is_wta:
            nengo.Connection(self.elem_output, self.elem_input,
                             synapse=inhibit_synapse,
                             transform=((np.eye(self.n_items) - 1) *
                                        inhibit_scale))
            self.is_wta = True
        else:
            warnings.warn("AssociativeMemory network is already configured "
                          "with a WTA network. Additional `add_wta_network` "
                          "calls are ignored.")

    @with_self
    def add_threshold_to_outputs(self, n_neurons=50, inhibit_scale=10):
        """Adds a thresholded output to the associative memory.

        Parameters
        ----------
        n_neurons: int, optional (Default: 50)
            Number of neurons to use for the default output vector ensemble.
        inhibit_scale: float, optional (Default: 10)
            Mutual inhibition scaling factor.
        """
        if self.thresh_ens is not None:
            warnings.warn("AssociativeMemory network is already configured "
                          "with thresholded outputs. Additional "
                          "`add_threshold_to_output` calls are ignored.")
            return

        with self.thresh_ens_config:
            self.thresh_bias = EnsembleArray(
                n_neurons, self.n_items, label='thresh_bias')
            self.thresh_ens = EnsembleArray(
                n_neurons, self.n_items, label='thresh_ens')
            self.thresholded_utilities = self.thresh_ens.output

            nengo.Connection(self.bias_node, self.thresh_bias.input,
                             transform=np.ones((self.n_items, 1)),
                             synapse=None)
            nengo.Connection(self.bias_node, self.thresh_ens.input,
                             transform=np.ones((self.n_items, 1)),
                             synapse=None)
            nengo.Connection(self.elem_output, self.thresh_bias.input,
                             transform=-inhibit_scale)
            nengo.Connection(self.thresh_bias.output, self.thresh_ens.input,
                             transform=-inhibit_scale)

            # Reroute connections from elem_output to be connections
            # from thresh_ens.output
            def reroute(conn):
                c = nengo.Connection(self.thresh_ens.output, conn.post,
                                     transform=conn.transform,
                                     synapse=conn.synapse)
                self.connections.remove(conn)
                return c
            self.default_vector_inhibit_conns = [
                reroute(c) for c in self.default_vector_inhibit_conns]
            self.out_conns = [reroute(c) for c in self.out_conns]

            # Make inhibitory connection if inhibit option is set
            if self.inhibit is not None:
                for e in self.thresh_ens.ensembles:
                    nengo.Connection(
                        self.inhibit, e, transform=-self._inhib_scale,
                        synapse=None)



import numpy as np
import pytest
from matplotlib.mlab import griddata

import nengo
from nengo.dists import UniformHypersphere
from nengo.utils.connection import target_function, eval_point_decoding
from nengo.utils.numpy import rms


@pytest.mark.parametrize("dimensions", [1, 4])
@pytest.mark.parametrize("radius", [1, 2.0])
@pytest.mark.filterwarnings("ignore:'targets' can be passed directly")
def test_target_function(Simulator, nl_nodirect, plt, dimensions, radius,
                         seed, rng):
    eval_points = UniformHypersphere().sample(1000, dimensions, rng=rng)
    eval_points *= radius
    f = lambda x: x ** 2
    targets = f(eval_points)

    model = nengo.Network(seed=seed)
    with model:
        model.config[nengo.Ensemble].neuron_type = nl_nodirect()
        inp = nengo.Node(lambda t: np.sin(t * 2 * np.pi) * radius)
        ens1 = nengo.Ensemble(40 * dimensions, dimensions, radius=radius)
        n1 = nengo.Node(size_in=dimensions)
        n2 = nengo.Node(size_in=dimensions)
        transform = np.linspace(1, -1, num=dimensions).reshape(-1, 1)
        nengo.Connection(inp, ens1, transform=transform)
        # pass in eval_points and targets
        nengo.Connection(ens1, n1, **target_function(eval_points, targets))
        # same, but let the builder apply f
        nengo.Connection(ens1, n2, function=f)
        probe1 = nengo.Probe(n1, synapse=0.03)
        probe2 = nengo.Probe(n2, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    plt.subplot(2, 1, 1)
    plt.plot(sim.trange(), sim.data[probe1])
    plt.title('Square manually with target_function')
    plt.subplot(2, 1, 2)
    plt.plot(sim.trange(), sim.data[probe2])
    plt.title('Square by passing in function to connection')

    assert np.allclose(sim.data[probe1], sim.data[probe2], atol=0.2 * radius)


def test_eval_point_decoding(Simulator, nl_nodirect, plt, seed):
    with nengo.Network(seed=seed) as model:
        model.config[nengo.Ensemble].neuron_type = nl_nodirect()
        a = nengo.Ensemble(200, 2)
        b = nengo.Ensemble(100, 1)
        c = nengo.Connection(a, b, function=lambda x: x[0] * x[1])

    with Simulator(model) as sim:
        eval_points, targets, decoded = eval_point_decoding(c, sim)

    def contour(xy, z):
        xi = np.linspace(-1, 1, 101)
        yi = np.linspace(-1, 1, 101)
        zi = griddata(xy[:, 0], xy[:, 1], z.ravel(), xi, yi, interp='linear')
        plt.contourf(xi, yi, zi, cmap=plt.cm.seismic)
        plt.colorbar()

    plt.figure(figsize=(15, 5))
    plt.subplot(131)
    contour(eval_points, targets)
    plt.title("Target (desired decoding)")
    plt.subplot(132)
    plt.title("Actual decoding")
    contour(eval_points, decoded)
    plt.subplot(133)
    plt.title("Difference between actual and desired")
    contour(eval_points, decoded - targets)

    # Generous error check, just to make sure it's in the right ballpark.
    # Also make sure error is above zero, i.e. y != z
    error = rms(decoded - targets, axis=1).mean()
    assert error < 0.1 and error > 1e-8



import errno
import multiprocessing
import os
import timeit
from subprocess import CalledProcessError

import numpy as np
from numpy.testing import assert_equal
import pytest

import nengo
from nengo.cache import (CacheIndex, DecoderCache, Fingerprint,
                         get_fragment_size, WriteableCacheIndex)
from nengo.exceptions import CacheIOWarning, FingerprintError
from nengo.solvers import LstsqL2
from nengo.utils.compat import int_types


class SolverMock(object):
    n_calls = {}

    def __init__(self):
        self.n_calls[self] = 0

    def __call__(self, conn, gain, bias, x, targets, rng=np.random, E=None):
        self.n_calls[self] += 1
        if E is None:
            return np.random.rand(x.shape[1], targets.shape[1]), {'info': 'v'}
        else:
            return np.random.rand(x.shape[1], E.shape[1]), {'info': 'v'}


def get_solver_test_args(**kwargs):
    M = 100
    N = 10
    D = 2
    conn = nengo.Connection(
        nengo.Ensemble(N, D, add_to_container=False),
        nengo.Node(size_in=D, add_to_container=False),
        add_to_container=False)
    conn.solver = kwargs.pop('solver', nengo.solvers.LstsqL2nz())
    defaults = {
        'conn': conn,
        'gain': np.ones(N),
        'bias': np.ones(N),
        'x': np.ones((M, D)),
        'targets': np.ones((M, N)),
        'rng': np.random.RandomState(42),
    }
    defaults.update(kwargs)
    return defaults


def get_weight_solver_test_args():
    M = 100
    N = 10
    N2 = 5
    D = 2
    conn = nengo.Connection(
        nengo.Ensemble(N, D, add_to_container=False),
        nengo.Node(size_in=D, add_to_container=False),
        solver=nengo.solvers.LstsqL2nz(), add_to_container=False)
    return {
        'conn': conn,
        'gain': np.ones(N),
        'bias': np.ones(N),
        'x': np.ones((M, D)),
        'targets': np.ones((M, N)),
        'rng': np.random.RandomState(42),
        'E': np.ones((D, N2)),
    }


def test_decoder_cache(tmpdir):
    cache_dir = str(tmpdir)

    # Basic test, that results are cached.
    with DecoderCache(cache_dir=cache_dir) as cache:
        solver_mock = SolverMock()
        decoders1, solver_info1 = cache.wrap_solver(solver_mock)(
            **get_solver_test_args())
        assert SolverMock.n_calls[solver_mock] == 1
        decoders2, solver_info2 = cache.wrap_solver(solver_mock)(
            **get_solver_test_args())
        assert SolverMock.n_calls[solver_mock] == 1  # result read from cache?
        assert_equal(decoders1, decoders2)
        assert solver_info1 == solver_info2

        solver_args = get_solver_test_args()
        solver_args['gain'] *= 2
        decoders3, solver_info3 = cache.wrap_solver(solver_mock)(**solver_args)
        assert SolverMock.n_calls[solver_mock] == 2
        assert np.any(decoders1 != decoders3)

        # Test that the cache does not load results of another solver.
        another_solver = SolverMock()
        cache.wrap_solver(another_solver)(**get_solver_test_args(
            solver=nengo.solvers.LstsqNoise()))
        assert SolverMock.n_calls[another_solver] == 1


def test_corrupted_decoder_cache(tmpdir):
    cache_dir = str(tmpdir)

    with DecoderCache(cache_dir=cache_dir) as cache:
        solver_mock = SolverMock()
        cache.wrap_solver(solver_mock)(**get_solver_test_args())
        assert SolverMock.n_calls[solver_mock] == 1

        # corrupt the cache
        for path in cache.get_files():
            with open(path, 'w') as f:
                f.write('corrupted')

        cache.wrap_solver(solver_mock)(**get_solver_test_args())
        assert SolverMock.n_calls[solver_mock] == 2


def test_corrupted_decoder_cache_index(tmpdir):
    cache_dir = str(tmpdir)

    with DecoderCache(cache_dir=cache_dir):
        pass  # Initialize cache with required files
    assert len(os.listdir(cache_dir)) == 2  # index, index.lock

    # Write corrupted index
    with open(os.path.join(cache_dir, CacheIndex._INDEX), 'w') as f:
        f.write('(d')  # empty dict, but missing '.' at the end

    # Try to load index
    with DecoderCache(cache_dir=cache_dir):
        pass
    assert len(os.listdir(cache_dir)) == 2  # index, index.lock


def test_decoder_cache_invalidation(tmpdir):
    cache_dir = str(tmpdir)
    solver_mock = SolverMock()

    # Basic test, that results are cached.
    with DecoderCache(cache_dir=cache_dir) as cache:
        cache.wrap_solver(solver_mock)(**get_solver_test_args())
        assert SolverMock.n_calls[solver_mock] == 1
        cache.invalidate()
        cache.wrap_solver(solver_mock)(**get_solver_test_args())
        assert SolverMock.n_calls[solver_mock] == 2


def test_decoder_cache_size_includes_overhead(tmpdir):
    cache_dir = str(tmpdir)
    solver_mock = SolverMock()

    with DecoderCache(cache_dir=cache_dir) as cache:
        cache.wrap_solver(solver_mock)(**get_solver_test_args())

        fragment_size = get_fragment_size(cache_dir)
        actual_size = sum(os.stat(p).st_size for p in cache.get_files())
        assert actual_size % fragment_size != 0, (
            'Test succeeded by chance. Adjust get_solver_test_args() to '
            'produce date not aligned with the files system fragment size.')

        assert cache.get_size_in_bytes() % fragment_size == 0


def test_decoder_cache_shrinking(tmpdir):
    cache_dir = str(tmpdir)
    solver_mock = SolverMock()
    another_solver = SolverMock()

    with DecoderCache(cache_dir=cache_dir) as cache:
        cache.wrap_solver(solver_mock)(**get_solver_test_args())

        # Ensure differing time stamps (depending on the file system the
        # timestamp resolution might be as bad as 1 day).
        for path in cache.get_files():
            timestamp = os.stat(path).st_atime
            timestamp -= 60 * 60 * 24 * 2  # 2 days
            os.utime(path, (timestamp, timestamp))

    with DecoderCache(cache_dir=cache_dir) as cache:
        cache.wrap_solver(another_solver)(**get_solver_test_args(
            solver=nengo.solvers.LstsqNoise()))

        cache_size = cache.get_size_in_bytes()
        assert cache_size > 0

        cache.shrink(cache_size - 1)

        # check that older cached result was removed
        assert SolverMock.n_calls[solver_mock] == 1
        cache.wrap_solver(another_solver)(**get_solver_test_args(
            solver=nengo.solvers.LstsqNoise()))
        cache.wrap_solver(solver_mock)(**get_solver_test_args())
        assert SolverMock.n_calls[solver_mock] == 2
        assert SolverMock.n_calls[another_solver] == 1


def test_decoder_cache_shrink_threadsafe(monkeypatch, tmpdir):
    """Tests that shrink handles files deleted by other processes."""
    cache_dir = str(tmpdir)
    solver_mock = SolverMock()

    with DecoderCache(cache_dir=cache_dir) as cache:
        cache.wrap_solver(solver_mock)(**get_solver_test_args())
        limit = cache.get_size()

        # Ensure differing time stamps (depending on the file system the
        # timestamp resolution might be as bad as 1 day).
        for filename in os.listdir(cache.cache_dir):
            path = os.path.join(cache.cache_dir, filename)
            timestamp = os.stat(path).st_atime
            timestamp -= 60 * 60 * 24 * 2  # 2 days
            os.utime(path, (timestamp, timestamp))

        cache.wrap_solver(solver_mock)(**get_solver_test_args(
            solver=nengo.solvers.LstsqNoise()))

        cache_size = cache.get_size_in_bytes()
        assert cache_size > 0

        def raise_file_not_found(orig_fn):
            def fn(filename, *args, **kwargs):
                if filename.endswith('.lock'):
                    return orig_fn(filename, *args, **kwargs)
                raise OSError(errno.ENOENT, "File not found.")
            return fn

        monkeypatch.setattr(cache, 'get_size_in_bytes', lambda: cache_size)
        monkeypatch.setattr('os.stat', raise_file_not_found(os.stat))
        monkeypatch.setattr('os.remove', raise_file_not_found(os.remove))
        monkeypatch.setattr('os.unlink', raise_file_not_found(os.unlink))

        cache.shrink(limit)


def test_decoder_cache_with_E_argument_to_solver(tmpdir):
    cache_dir = str(tmpdir)
    solver_mock = SolverMock()

    with DecoderCache(cache_dir=cache_dir) as cache:
        decoders1, solver_info1 = cache.wrap_solver(solver_mock)(
            **get_weight_solver_test_args())
        assert SolverMock.n_calls[solver_mock] == 1
        decoders2, solver_info2 = cache.wrap_solver(solver_mock)(
            **get_weight_solver_test_args())
        assert SolverMock.n_calls[solver_mock] == 1  # read from cache?
        assert_equal(decoders1, decoders2)
        assert solver_info1 == solver_info2


class DummyA(object):
    def __init__(self, attr=0):
        self.attr = attr


nengo.cache.Fingerprint.whitelist(DummyA)


class DummyB(object):
    def __init__(self, attr=0):
        self.attr = attr


nengo.cache.Fingerprint.whitelist(DummyB)


def dummy_fn(arg):
    pass


@pytest.mark.parametrize('reference, equal, different', (
    (True, True, False),             # bool
    (False, False, True),            # bool
    (1.0, 1.0, 2.0),                 # float
    (1.0 + 2.0j, 1 + 2j, 2.0 + 1j),  # complex
    (b'a', b'a', b'b'),              # bytes
    ((0, 1), (0, 1), (0, 2)),        # tuple
    ([0, 1], [0, 1], [0, 2]),        # list
    (u'a', u'a', u'b'),              # unicode string
    (np.eye(2), np.eye(2), np.array([[0, 1], [1, 0]])),      # array
    (DummyA(), DummyA(), DummyB()),  # object instance
    (DummyA(1), DummyA(1), DummyA(2)),     # object instance
    (LstsqL2(reg=.1), LstsqL2(reg=.1), LstsqL2(reg=.2)),     # solver
) + tuple((typ(1), typ(1), typ(2)) for typ in int_types))
def test_fingerprinting(reference, equal, different):
    assert str(Fingerprint(reference)) == str(Fingerprint(equal))
    assert str(Fingerprint(reference)) != str(Fingerprint(different))


@pytest.mark.parametrize('obj', (
    np.array([object()]),   # array
    np.array([(1.,)], dtype=[('field1', 'f8')]),  # array
    {'a': 1, 'b': 2},       # dict
    object(),               # object instance
    dummy_fn,               # function
))
def test_unsupported_fingerprinting(obj):
    with pytest.raises(FingerprintError):
        Fingerprint(obj)


def test_fails_for_lambda_expression():
    with pytest.raises(FingerprintError):
        Fingerprint(lambda x: x)


def test_cache_works(tmpdir, RefSimulator, seed):
    cache_dir = str(tmpdir)

    model = nengo.Network(seed=seed)
    with model:
        nengo.Connection(nengo.Ensemble(10, 1), nengo.Ensemble(10, 1))

    assert len(os.listdir(cache_dir)) == 0
    with RefSimulator(model, model=nengo.builder.Model(
            dt=0.001, decoder_cache=DecoderCache(cache_dir=cache_dir))):
        assert len(os.listdir(cache_dir)) == 3  # index, index.lock, and *.nco


def test_cache_not_used_without_seed(tmpdir, RefSimulator):
    cache_dir = str(tmpdir)

    model = nengo.Network()
    with model:
        nengo.Connection(nengo.Ensemble(10, 1), nengo.Ensemble(10, 1))

    assert len(os.listdir(cache_dir)) == 0
    with RefSimulator(model, model=nengo.builder.Model(
            dt=0.001, decoder_cache=DecoderCache(cache_dir=cache_dir))):
        assert len(os.listdir(cache_dir)) == 2  # index, index.lock


def build_many_ensembles(cache_dir, RefSimulator):
    with nengo.Network(seed=1) as model:
        for _ in range(100):
            nengo.Connection(nengo.Ensemble(10, 1), nengo.Ensemble(10, 1))

    with RefSimulator(model, model=nengo.builder.Model(
            dt=0.001, decoder_cache=DecoderCache(cache_dir=cache_dir))):
        pass


@pytest.mark.slow
def test_cache_concurrency(tmpdir, RefSimulator):
    cache_dir = str(tmpdir)

    n_processes = 100
    processes = [
        multiprocessing.Process(
            target=build_many_ensembles, args=(cache_dir, RefSimulator))
        for _ in range(n_processes)]
    for p in processes:
        p.start()
    for p in processes:
        p.join(60)
    for p in processes:
        assert p.exitcode == 0


def reject_outliers(data):
    med = np.median(data)
    limits = 1.5 * (np.percentile(data, [25, 75]) - med) + med
    return np.asarray(data)[np.logical_and(data > limits[0], data < limits[1])]


class TestCacheBenchmark(object):
    n_trials = 25

    setup = '''
import numpy as np
import nengo
import nengo.cache
from nengo.rc import rc

rc.set("decoder_cache", "path", {tmpdir!r})

model = nengo.Network(seed=1)
with model:
    a = nengo.Ensemble({N}, dimensions={D}, n_eval_points={M})
    b = nengo.Ensemble({N}, dimensions={D}, n_eval_points={M})
    conn = nengo.Connection(a, b)
    '''

    without_cache = {
        'rc': 'rc.set("decoder_cache", "enabled", "False")',
        'stmt': '''
with nengo.Simulator(model):
    pass
'''
    }

    with_cache_miss_ro = {
        'rc': '''
with nengo.cache.DecoderCache() as cache:
    cache.invalidate()
rc.set("decoder_cache", "enabled", "True")
rc.set("decoder_cache", "readonly", "True")
''',
        'stmt': '''
with nengo.Simulator(model):
    pass
'''
    }

    with_cache_miss = {
        'rc': '''
with nengo.cache.DecoderCache() as cache:
    cache.invalidate()
rc.set("decoder_cache", "enabled", "True")
rc.set("decoder_cache", "readonly", "False")
''',
        'stmt': '''
with nengo.Simulator(model):
    pass
'''
    }

    with_cache_hit = {
        'rc': '''
rc.set("decoder_cache", "enabled", "True")
rc.set("decoder_cache", "readonly", "False")
with nengo.Simulator(model):
    pass
''',
        'stmt': '''
with nengo.Simulator(model):
    pass
'''
    }

    labels = ["no cache", "cache miss", "cache miss ro", "cache hit"]
    keys = [l.replace(' ', '_') for l in labels]
    param_to_axis_label = {
        'D': "dimensions",
        'N': "neurons",
        'M': "evaluation points"
    }
    defaults = {'D': 1, 'N': 50, 'M': 1000}

    def time_code(self, code, **kwargs):
        return timeit.repeat(
            stmt=code['stmt'], setup=self.setup.format(**kwargs) + code['rc'],
            number=1, repeat=self.n_trials)

    def time_all(self, **kwargs):
        return (self.time_code(self.without_cache, **kwargs),
                self.time_code(self.with_cache_miss, **kwargs),
                self.time_code(self.with_cache_miss_ro, **kwargs),
                self.time_code(self.with_cache_hit, **kwargs))

    def get_args(self, varying_param, value):
        args = dict(self.defaults)  # make a copy
        args[varying_param] = value
        return args

    @pytest.mark.slow
    @pytest.mark.noassertions
    @pytest.mark.parametrize('varying_param', ['D', 'N', 'M'])
    def test_cache_benchmark(self, tmpdir, varying_param, analytics, plt):
        varying = {
            'D': np.asarray(np.linspace(1, 512, 10), dtype=int),
            'N': np.asarray(np.linspace(10, 500, 8), dtype=int),
            'M': np.asarray(np.linspace(750, 2500, 8), dtype=int)
        }[varying_param]
        axis_label = self.param_to_axis_label[varying_param]

        times = [
            self.time_all(
                tmpdir=str(tmpdir), **self.get_args(varying_param, v))
            for v in varying]

        for i, data in enumerate(zip(*times)):
            plt.plot(varying, np.median(data, axis=1), label=self.labels[i])
            plt.xlim(right=varying[-1])
            analytics.add_data(varying_param, varying, axis_label)
            analytics.add_data(self.keys[i], data)

        plt.xlabel("Number of %s" % axis_label)
        plt.ylabel("Build time (s)")
        plt.legend(loc='best')

    @pytest.mark.compare
    @pytest.mark.parametrize('varying_param', ['D', 'N', 'M'])
    def test_compare_cache_benchmark(
            self, varying_param, analytics_data, plt, logger):
        stats = pytest.importorskip('scipy.stats')

        d1, d2 = analytics_data
        assert np.all(d1[varying_param] == d2[varying_param]), (
            'Cannot compare different parametrizations')
        axis_label = self.param_to_axis_label[varying_param]

        logger.info("Cache, varying %s:", axis_label)
        for label, key in zip(self.labels, self.keys):
            clean_d1 = [reject_outliers(d) for d in d1[key]]
            clean_d2 = [reject_outliers(d) for d in d2[key]]
            diff = [np.median(b) - np.median(a)
                    for a, b in zip(clean_d1, clean_d2)]

            p_values = np.array(
                [2. * stats.mannwhitneyu(a, b, alternative='two-sided')[1]
                 for a, b in zip(clean_d1, clean_d2)])
            overall_p = 1. - np.prod(1. - p_values)
            if overall_p < .05:
                logger.info("  %s: Significant change (p <= %.3f). See plots.",
                            label, np.ceil(overall_p * 1000.) / 1000.)
            else:
                logger.info("  %s: No significant change.", label)

            plt.plot(d1[varying_param], diff, label=label)

        plt.xlabel("Number of %s" % axis_label)
        plt.ylabel("Difference in build time (s)")
        plt.legend(loc='best')


class TestCacheShrinkBenchmark(object):
    n_trials = 50

    setup = '''
import numpy as np
import nengo
import nengo.cache
from nengo.rc import rc

rc.set("decoder_cache", "path", {tmpdir!r})

for i in range(10):
    model = nengo.Network(seed=i)
    with model:
        a = nengo.networks.EnsembleArray(10, 128, 1)
        b = nengo.networks.EnsembleArray(10, 128, 1)
        conn = nengo.Connection(a.output, b.input)
    with nengo.Simulator(model):
        pass

rc.set("decoder_cache", "size", "0KB")
cache = nengo.cache.DecoderCache()
    '''

    stmt = 'with cache: cache.shrink()'

    @pytest.mark.slow
    @pytest.mark.noassertions
    def test_cache_shrink_benchmark(self, tmpdir, analytics, logger):
        times = timeit.repeat(
            stmt=self.stmt, setup=self.setup.format(tmpdir=str(tmpdir)),
            number=1, repeat=self.n_trials)
        logger.info("Shrink took a minimum of %f seconds.", np.min(times))
        logger.info("Shrink took a %f seconds on average.", np.mean(times))
        logger.info(
            "Shrink took a %f seconds on average with outliers rejected.",
            np.mean(reject_outliers(times)))
        analytics.add_data('times', times)

    @pytest.mark.compare
    def test_compare_cache_shrink_benchmark(self, analytics_data, plt, logger):
        stats = pytest.importorskip('scipy.stats')

        d1, d2 = (x['times'] for x in analytics_data)
        clean_d1 = reject_outliers(d1)
        clean_d2 = reject_outliers(d2)

        diff = np.median(clean_d2) - np.median(clean_d1)

        p_value = 2. * stats.mannwhitneyu(
            clean_d1, clean_d2, alternative='two-sided')[1]
        if p_value < .05:
            logger.info("Significant change of %d seconds (p <= %.3f).",
                        diff, np.ceil(p_value * 1000.) / 1000.)
        else:
            logger.info("No significant change (%d).", diff)
        logger.info("Speed up: %s", np.median(clean_d1) / np.median(clean_d2))

        plt.scatter(np.ones_like(d1), d1, c='b')
        plt.scatter(2 * np.ones_like(d2), d2, c='g')


def test_warns_out_of_context(tmpdir):
    cache_dir = str(tmpdir)
    cache = DecoderCache(cache_dir=cache_dir)

    solver_mock = SolverMock()
    solver = cache.wrap_solver(solver_mock)
    with pytest.warns(UserWarning):
        solver(**get_solver_test_args())
    assert SolverMock.n_calls[solver_mock] == 1


def test_cacheindex_cannot_write(tmpdir):
    index = WriteableCacheIndex(cache_dir=str(tmpdir))
    with index:
        index[0] = ("file0", 0, 0)
    mtime = os.stat(index.index_path).st_mtime

    index = CacheIndex(cache_dir=str(tmpdir))
    with index:
        with pytest.raises(TypeError):
            index[0] = ("file", 0, 0)
        with pytest.raises(TypeError):
            del index[0]
        assert index[0] == ("file0", 0, 0)
    assert os.stat(index.index_path).st_mtime == mtime


def test_writeablecacheindex_writes(tmpdir):
    index = WriteableCacheIndex(cache_dir=str(tmpdir))
    with index:
        index[0] = ("file0", 0, 0)
        index[1] = ("file1", 0, 0)
        del index[1]

    # Verify with readonly cacheindex
    index = CacheIndex(cache_dir=str(tmpdir))
    with index:
        assert index[0] == ("file0", 0, 0)
        assert 1 not in index


def test_writeablecacheindex_setitem(tmpdir):
    index = WriteableCacheIndex(cache_dir=str(tmpdir))

    with pytest.raises(ValueError):
        index[0] = "file0"
    with pytest.raises(ValueError):
        index[0] = ("file0", 0)
    with pytest.raises(ValueError):
        index[0] = ("file0", 0, 0, 0)


def test_writeablecacheindex_removes(tmpdir):
    index = WriteableCacheIndex(cache_dir=str(tmpdir))
    with index:
        index[0] = ("file0", 0, 0)
        index[1] = ("file1", 0, 0)
        index.remove_file_entry("file0")
        index.remove_file_entry(os.path.join(str(tmpdir), "file1"))

    # Verify with readonly cacheindex
    index = CacheIndex(cache_dir=str(tmpdir))
    with index:
        assert 0 not in index, "Fails on relative paths"
        assert 1 not in index, "Fails on absolute paths"


def test_writeablecacheindex_warning(monkeypatch, tmpdir):

    def raise_error(*args, **kwargs):
        raise CalledProcessError(-1, "move")

    monkeypatch.setattr(nengo.cache, "replace", raise_error)
    with pytest.warns(CacheIOWarning):
        with WriteableCacheIndex(cache_dir=str(tmpdir)):
            pass


def test_shrink_does_not_fail_if_lock_cannot_be_acquired(tmpdir):
    cache = DecoderCache(cache_dir=str(tmpdir))
    cache._index._lock.timeout = 1.
    with cache:
        cache.wrap_solver(SolverMock())(**get_solver_test_args())
    with cache._index._lock:
        cache.shrink(limit=0)



import numpy as np
import pytest

import nengo

from nengolib.networks.linear_network import LinearNetwork
from nengolib import Network, Lowpass
from nengolib.signal import s, z, canonical, Identity, shift, nrmse
from nengolib.synapses import PadeDelay, Bandpass
from nengolib.testing import warns


_mock_solver_calls = 0  # global to keep solver's fingerprint static


class MockSolver(nengo.solvers.LstsqL2):

    def __call__(self, A, Y, __hack__=None, **kwargs):
        assert __hack__ is None
        # __hack__ is necessary prior to nengo PR #1359 (<2.6.1)
        # and following nengo PR #1507 (>2.8.0)

        globals()['_mock_solver_calls'] += 1
        return super(MockSolver, self).__call__(A, Y, **kwargs)


@pytest.mark.parametrize("neuron_type,atol,atol_x", [
    (nengo.neurons.Direct(), 1e-14, 1e-14),
    (nengo.neurons.LIFRate(), 0.01, 0.01),
    (nengo.neurons.LIF(), 0.1, 0.1)])
def test_linear_network(Simulator, plt, seed, rng, neuron_type, atol, atol_x):
    n_neurons = 500
    dt = 0.001
    T = 1.0

    sys = Lowpass(0.1)
    scale_input = 2.0

    synapse = 0.02
    tau_probe = 0.005

    with Network(seed=seed) as model:
        stim = nengo.Node(
            output=nengo.processes.WhiteSignal(T, high=10, seed=seed))
        subnet = LinearNetwork(
            sys, n_neurons_per_ensemble=n_neurons, synapse=synapse,
            input_synapse=synapse, dt=dt, neuron_type=neuron_type)
        nengo.Connection(
            stim, subnet.input, synapse=None, transform=scale_input)

        assert subnet.synapse == subnet.input_synapse
        assert subnet.output_synapse is None

        p_input = nengo.Probe(subnet.input, synapse=tau_probe)
        p_x = nengo.Probe(subnet.state.output, synapse=tau_probe)
        p_output = nengo.Probe(subnet.output, synapse=tau_probe)

    with Simulator(model, dt=dt) as sim:
        sim.run(T)

    ideal_output = shift(sys.filt(sim.data[p_input]))
    ideal_x = shift(subnet.realization.X.filt(sim.data[p_input]))

    plt.plot(sim.trange(), sim.data[p_input], label="Input", alpha=0.5)
    plt.plot(sim.trange(), sim.data[p_output], label="Actual y", alpha=0.5)
    plt.plot(sim.trange(), ideal_output, label="Expected y", alpha=0.5,
             linestyle='--')
    plt.plot(sim.trange(), sim.data[p_x], label="Actual x", alpha=0.5)
    plt.plot(sim.trange(), ideal_x, label="Expected x", alpha=0.5,
             linestyle='--')
    plt.legend()

    assert nrmse(sim.data[p_output], ideal_output) < atol
    assert nrmse(sim.data[p_x].squeeze(), ideal_x.squeeze()) < atol_x


def test_none_dt(Simulator, seed, rng):
    dt = 0.0001  # approximates continuous case
    T = 1.0

    sys = Bandpass(8, 5)
    synapse = 0.01

    with Network(seed=seed) as model:
        stim = nengo.Node(
            output=nengo.processes.WhiteSignal(T, high=10, seed=seed))
        subnet = LinearNetwork(
            sys, 1, synapse=synapse, input_synapse=synapse, dt=None,
            neuron_type=nengo.neurons.Direct())
        nengo.Connection(stim, subnet.input, synapse=None)

        assert subnet.output_synapse is None

        p_ideal = nengo.Probe(subnet.input, synapse=sys)
        p_output = nengo.Probe(subnet.output, synapse=None)

    with Simulator(model, dt=dt) as sim:
        sim.run(T)

    assert np.allclose(sim.data[p_output], sim.data[p_ideal], atol=0.1)


def test_output_filter(Simulator, seed, rng):
    dt = 0.001
    T = 1.0

    sys = PadeDelay(0.1, order=3, p=3)
    assert sys.has_passthrough
    synapse = 0.01

    with Network(seed=seed) as model:
        stim = nengo.Node(
            output=nengo.processes.WhiteSignal(T, high=10, seed=seed))
        subnet = LinearNetwork(
            sys, 1, synapse=synapse, output_synapse=synapse, dt=dt,
            neuron_type=nengo.neurons.Direct())
        nengo.Connection(stim, subnet.input, synapse=None)

        assert subnet.input_synapse is None

        p_ideal = nengo.Probe(subnet.input, synapse=sys)
        p_output = nengo.Probe(subnet.output, synapse=None)

    with Simulator(model, dt=dt) as sim:
        sim.run(T)

    assert np.allclose(sim.data[p_output][:-1], sim.data[p_ideal][1:])


def test_unfiltered(Simulator, seed, rng):
    dt = 0.001
    T = 1.0

    sys = nengo.Alpha(0.1)
    synapse = 0.01

    with Network(seed=seed) as model:
        stim = nengo.Node(
            output=nengo.processes.WhiteSignal(T, high=10, seed=seed))
        subnet = LinearNetwork(
            sys, 1, synapse=synapse, dt=dt,
            neuron_type=nengo.neurons.Direct())
        nengo.Connection(stim, subnet.input, synapse=None)

        assert subnet.input_synapse is None
        assert subnet.output_synapse is None

        p_ideal = nengo.Probe(subnet.input, synapse=sys)
        p_output = nengo.Probe(subnet.output, synapse=synapse)

    with Simulator(model, dt=dt) as sim:
        sim.run(T)

    assert np.allclose(sim.data[p_output], sim.data[p_ideal])


def test_unstable_warning():
    with warns(UserWarning):
        with Network():
            LinearNetwork(~s, 1, synapse=0.02, dt=0.001, realizer=Identity())


def test_output_warning():
    with warns(UserWarning):
        LinearNetwork(([1, 1], [1, 1]), 1, synapse=1, dt=1)


def test_invalid_systems():
    with pytest.raises(ValueError):
        LinearNetwork(~z, 1, synapse=1, dt=1)

    with pytest.raises(ValueError):
        LinearNetwork(~s, 1, synapse=~z, dt=1)

    with pytest.raises(ValueError):
        LinearNetwork(1, 1, synapse=1, dt=1)


def test_radii(Simulator, seed, plt):
    sys = canonical(PadeDelay(0.2, order=3))
    dt = 0.001
    T = 0.5

    plt.figure()

    # Precompute the exact bounds for an impulse stimulus
    radii = []
    for sub in sys:
        response = sub.impulse(int(T / dt), dt=dt)
        amplitude = np.max(abs(response))
        assert amplitude >= 1e-4  # otherwise numerical issues
        radii.append(amplitude)

        plt.plot(response / amplitude, linestyle='--')

    with Network(seed=seed) as model:
        # Impulse stimulus
        stim = nengo.Node(output=lambda t: 1 / dt if t <= dt else 0)

        # Set explicit radii for controllable realization
        subnet = LinearNetwork(sys, n_neurons_per_ensemble=1,
                               synapse=0.2, input_synapse=0.2, dt=dt,
                               radii=radii, realizer=Identity(),
                               neuron_type=nengo.neurons.Direct())
        nengo.Connection(stim, subnet.input, synapse=None)
        p = nengo.Probe(subnet.state.output, synapse=None)

    with Simulator(model, dt=dt) as sim:
        sim.run(T)

    plt.plot(sim.data[p], lw=5, alpha=0.5)

    assert np.allclose(np.max(abs(sim.data[p]), axis=0), 1, atol=1e-4)


def test_solver(tmpdir, Simulator, seed, rng):
    assert _mock_solver_calls == 0

    for _ in range(3):
        model = LinearNetwork(
            nengo.Lowpass(0.1), 10, synapse=0.1, dt=0.001,
            solver=MockSolver(reg=rng.rand()), seed=seed)

        Simulator(model, model=nengo.builder.Model(
            decoder_cache=nengo.cache.DecoderCache(cache_dir=str(tmpdir))))

    # checks workaround for https://github.com/nengo/nengo/issues/1044
    assert _mock_solver_calls == 3



import numpy as np

import nengo
from nengo.utils.builder import (
    generate_graphviz, remove_passthrough_nodes, objs_and_connections)


def test_create_dot():
    """Constructing a .dot file for a model"""

    model = nengo.Network()
    with model:
        D = 3
        input = nengo.Node([1] * D, label='input')
        a = nengo.networks.EnsembleArray(50, D, label='a')
        b = nengo.networks.EnsembleArray(50, D, label='b')
        output = nengo.Node(None, size_in=D, label='output')

        nengo.Connection(input, a.input, synapse=0.01)
        nengo.Connection(a.output, b.input, synapse=0.01)
        nengo.Connection(b.output, b.input, synapse=0.01, transform=0.9)
        nengo.Connection(a.output, a.input, synapse=0.01,
                         transform=np.ones((D, D)))
        nengo.Connection(b.output, output, synapse=0.01)

    objs, conns = objs_and_connections(model)
    dot = generate_graphviz(objs, conns)
    assert len(dot.splitlines()) == 31
    # not sure what else to check here

    dot = generate_graphviz(
        *remove_passthrough_nodes(objs, conns))
    assert len(dot.splitlines()) == 27
    # not sure what else to check here



import numpy as np
import pytest

import nengo

from nengolib.synapses.mapping import ss2sim
from nengolib import Network, Lowpass, Alpha
from nengolib.signal import s, z, ss_equal, cont2discrete, shift
from nengolib.synapses import Highpass, PadeDelay, DoubleExp


def test_mapping(Simulator, plt, seed):
    sys = Alpha(0.1)
    syn = Lowpass(0.01)
    gsyn = 2*syn  # scaled lowpass
    isyn = 2/s  # scaled integrator
    dt = 0.001

    ss = ss2sim(sys, syn, None)  # normal lowpass, continuous
    dss = ss2sim(sys, syn, dt)  # normal lowpass, discrete
    gss = ss2sim(sys, gsyn, None)  # scaled lowpass, continuous
    gdss = ss2sim(sys, gsyn, dt)  # scaled lowpass, discrete
    iss = ss2sim(sys, isyn, None)  # scaled integrator, continuous
    idss = ss2sim(sys, isyn, dt)  # scaled integrator, discrete
    assert ss.analog and gss.analog and iss.analog
    assert not (dss.analog or gdss.analog or idss.analog)

    with Network(seed=seed) as model:
        stim = nengo.Node(output=lambda t: np.sin(20*np.pi*t))

        probes = []
        for mapped, synapse in ((ss, syn), (dss, syn), (gss, gsyn),
                                (gdss, gsyn), (iss, isyn), (idss, isyn)):
            A, B, C, D = mapped.ss
            x = nengo.Node(size_in=2)
            y = nengo.Node(size_in=1)

            nengo.Connection(stim, x, transform=B, synapse=synapse)
            nengo.Connection(x, x, transform=A, synapse=synapse)
            nengo.Connection(x, y, transform=C, synapse=None)
            nengo.Connection(stim, y, transform=D, synapse=None)

            probes.append(nengo.Probe(y))

        p_stim = nengo.Probe(stim)

    pss, pdss, pgss, pgdss, piss, pidss = probes

    with Simulator(model, dt=dt) as sim:
        sim.run(1.0)

    expected = shift(sys.filt(sim.data[p_stim], dt))

    plt.plot(sim.trange(), sim.data[pss], label="Continuous", alpha=0.5)
    plt.plot(sim.trange(), sim.data[pdss], label="Discrete", alpha=0.5)
    plt.plot(sim.trange(), sim.data[pgss], label="Gain Cont.", alpha=0.5)
    plt.plot(sim.trange(), sim.data[pgdss], label="Gain Disc.", alpha=0.5)
    plt.plot(sim.trange(), sim.data[piss], label="Integ Cont.", alpha=0.5)
    plt.plot(sim.trange(), sim.data[pidss], label="Integ Disc.", alpha=0.5)
    plt.plot(sim.trange(), expected, label="Expected", linestyle='--')
    plt.legend()

    assert np.allclose(sim.data[pss], expected, atol=0.01)
    assert np.allclose(sim.data[pdss], expected)
    assert np.allclose(sim.data[pgss], expected, atol=0.01)
    assert np.allclose(sim.data[pgdss], expected)
    assert np.allclose(sim.data[piss], expected, atol=0.01)
    assert np.allclose(sim.data[pidss], expected)


def test_principle3_continuous():
    sys = PadeDelay(0.1, order=5)

    tau = 0.01
    syn = Lowpass(tau)

    FH = ss2sim(sys, syn, dt=None)

    assert np.allclose(FH.A, tau * sys.A + np.eye(len(sys)))
    assert np.allclose(FH.B, tau * sys.B)
    assert np.allclose(FH.C, sys.C)
    assert np.allclose(FH.D, sys.D)


def test_principle3_discrete():
    sys = PadeDelay(0.1, order=5)

    tau = 0.01
    dt = 0.002
    syn = Lowpass(tau)

    FH = ss2sim(sys, syn, dt=dt)

    a = np.exp(-dt / tau)
    sys = cont2discrete(sys, dt=dt)
    assert np.allclose(FH.A, (sys.A - a * np.eye(len(sys))) / (1 - a))
    assert np.allclose(FH.B, sys.B / (1 - a))
    assert np.allclose(FH.C, sys.C)
    assert np.allclose(FH.D, sys.D)

    # We can also do the discretization ourselves and then pass in dt=None
    assert ss_equal(
        ss2sim(sys, cont2discrete(syn, dt=dt), dt=None), FH)


@pytest.mark.parametrize("sys", [PadeDelay(0.1, order=5), Lowpass(0.1)])
def test_doubleexp_continuous(sys):
    tau1 = 0.05
    tau2 = 0.02
    syn = DoubleExp(tau1, tau2)

    FH = ss2sim(sys, syn, dt=None)

    A = sys.A
    FHA = tau1 * tau2 * np.dot(A, A) + (tau1 + tau2) * A + np.eye(len(A))
    B = sys.B
    FHB = (tau1 * tau2 * A + (tau1 + tau2) * np.eye(len(A))).dot(B)
    assert np.allclose(FH.A, FHA)
    assert np.allclose(FH.B, FHB)
    assert np.allclose(FH.C, sys.C)
    assert np.allclose(FH.D, sys.D)


def test_doubleexp_discrete():
    sys = PadeDelay(0.1, order=5)

    tau1 = 0.05
    tau2 = 0.02
    dt = 0.002
    syn = DoubleExp(tau1, tau2)

    FH = ss2sim(sys, syn, dt=dt)

    a1 = np.exp(-dt / tau1)
    a2 = np.exp(-dt / tau2)
    t1 = 1 / (1 - a1)
    t2 = 1 / (1 - a2)
    c = [a1 * a2 * t1 * t2, - (a1 + a2) * t1 * t2, t1 * t2]
    sys = cont2discrete(sys, dt=dt)

    A = sys.A
    FHA = c[2] * np.dot(A, A) + c[1] * A + c[0] * np.eye(len(A))
    B = sys.B
    FHB = (c[2] * A + (c[1] + c[2]) * np.eye(len(A))).dot(B)
    assert np.allclose(FH.A, FHA)
    assert np.allclose(FH.B, FHB)
    assert np.allclose(FH.C, sys.C)
    assert np.allclose(FH.D, sys.D)


def test_unsupported_mapping():
    lpf = Lowpass(0.1)

    with pytest.raises(ValueError):
        ss2sim(sys=lpf, synapse=Highpass(0.1), dt=None)

    with pytest.raises(ValueError):
        ss2sim(sys=~z, synapse=lpf, dt=None)

    with pytest.raises(ValueError):
        ss2sim(sys=lpf, synapse=~z, dt=None)

    with pytest.raises(ValueError):
        ss2sim(sys=~z, synapse=~z, dt=1.)



import pytest

from nengo.exceptions import SpaParseError
from nengo.spa.action_objects import Symbol, Source, DotProduct

A = Symbol('A')
B = Symbol('B')
x = Source('x')
y = Source('y')


def test_symbol():
    A = Symbol('A')
    assert A.symbol == 'A'
    B = Symbol('B')
    assert str(B) == 'B'

    assert str(A + B) == '(A + B)'
    with pytest.raises(TypeError):
        A + "A"

    assert str(A - B) == '(A - B)'
    with pytest.raises(TypeError):
        A - []

    assert str(A * B) == '(A * B)'
    assert str(A * 1) == 'A'
    assert str(1 * A) == 'A'
    assert str(1.0 * A) == 'A'
    assert str(A * 1.0) == 'A'
    assert str(A * 1.5) == '(A * 1.5)'
    assert str(A * 2) == '(A * 2)'
    assert str(A * Symbol('1')) == 'A'
    assert str(Symbol('1') * A) == 'A'

    with pytest.raises(TypeError):
        A * {}

    assert str(~A) == '~A'
    assert str(~(~A)) == 'A'

    assert str(-A) == '-A'
    assert str(-(-A)) == 'A'

    assert str(A*B+(B*~A)*0.5-A) == '(((A * B) + ((B * ~A) * 0.5)) - A)'


def test_source():
    A = Symbol('A')
    s = Source('s')
    assert s.name == 's'
    assert str(s.transform) == '1'
    assert str(s) == 's'
    s_q = Source('s', Symbol('Q'))
    assert s_q.name == 's'
    assert str(s_q.transform) == 'Q'
    assert str(s_q) == 'Q * s'

    assert str(s * 1.5) == '1.5 * s'
    assert str(s * A) == 'A * s'
    assert str(s_q * A) == '(Q * A) * s'

    assert str(1.5 * s) == '1.5 * s'
    assert str(A * s) == 'A * s'
    assert str(A * s_q) == '(Q * A) * s'

    assert str(-s) == '-1 * s'
    assert str(-(-s)) == 's'
    assert str(-s_q) == '-Q * s'

    assert str(s + s_q) == 's + Q * s'

    with pytest.raises(TypeError):
        s * "A"


def test_inverted():
    assert str(~x) == '~x'
    assert str(A*~x) == 'A * ~x'
    assert str(~x*y) == '((~x) * (y)) * 1'
    assert str(~x*~y) == '((~x) * (~y)) * 1'
    assert str(x*~y) == '((x) * (~y)) * 1'
    with pytest.raises(ValueError):
        ~(2*x)
    with pytest.raises(ValueError):
        ~(A*x)


def test_dotproduct():
    assert str(DotProduct(x, y)) == 'dot(x, y)'
    assert str(DotProduct(x, A)) == 'dot(x, A)'
    assert str(DotProduct(B, y)) == 'dot(B, y)'

    with pytest.raises(SpaParseError):
        DotProduct(B, A)
    with pytest.raises(SpaParseError):
        DotProduct(A, "B")
    with pytest.raises(SpaParseError):
        DotProduct(1, A)

    assert str(0.5*DotProduct(x, y)) == '0.5 * dot(x, y)'
    assert str(0.5 * (0.5*DotProduct(x, y))) == '0.25 * dot(x, y)'
    assert str(DotProduct(B, y)*2) == '2 * dot(B, y)'

    with pytest.raises(TypeError):
        A * DotProduct(x, y)
    with pytest.raises(TypeError):
        DotProduct(x, y) * B

    assert str(DotProduct(x, y)/2) == '0.5 * dot(x, y)'
    assert str((DotProduct(x, y)/2)/2) == '0.25 * dot(x, y)'
    assert str(2*DotProduct(B, y)/2) == 'dot(B, y)'

    with pytest.raises(TypeError):
        DotProduct(x, y) / B

    assert str(DotProduct(x, y) + 1) == 'dot(x, y) + 1'
    assert str(1 + DotProduct(x, y)) == 'dot(x, y) + 1'
    assert str(DotProduct(x, y) + DotProduct(B, y)) == 'dot(x, y) + dot(B, y)'
    assert str(DotProduct(x, y) + A) == 'dot(x, y) + A'
    assert str(B + DotProduct(x, y)) == 'dot(x, y) + B'

    assert str(-DotProduct(x, y)) == '-dot(x, y)'
    assert str(-(-DotProduct(x, y))) == 'dot(x, y)'

    assert str(DotProduct(x, y) - 1) == 'dot(x, y) + -1'
    assert str(1 - DotProduct(x, y)) == '-dot(x, y) + 1'
    assert str(DotProduct(x, y) - DotProduct(B, y)) == 'dot(x, y) + -dot(B, y)'
    assert str(DotProduct(x, y) - A) == 'dot(x, y) + -A'
    assert str(B - DotProduct(x, y)) == '-dot(x, y) + B'


def test_condition_list():
    xy = DotProduct(x, y)
    Ay = DotProduct(A, y)

    assert str(xy + Ay) == 'dot(x, y) + dot(A, y)'

    assert str(2*(xy + Ay)) == '2 * dot(x, y) + 2 * dot(A, y)'
    assert str((xy + Ay)*2) == '2 * dot(x, y) + 2 * dot(A, y)'
    assert str(2*(xy + Ay)*2) == '4 * dot(x, y) + 4 * dot(A, y)'

    assert str(2*(xy + Ay)/2) == 'dot(x, y) + dot(A, y)'
    assert str(2*(xy + Ay)/0.5) == '4 * dot(x, y) + 4 * dot(A, y)'

    assert str(xy + Ay + xy) == 'dot(x, y) + dot(A, y) + dot(x, y)'
    assert str((xy + Ay) + xy) == 'dot(x, y) + dot(A, y) + dot(x, y)'
    assert str(2*(xy + Ay) + xy) == '2 * dot(x, y) + 2 * dot(A, y) + dot(x, y)'
    assert str(xy + (Ay + xy)) == 'dot(x, y) + dot(A, y) + dot(x, y)'

    assert str(xy + (Ay - xy)) == 'dot(x, y) + dot(A, y) + -dot(x, y)'
    assert str((xy + Ay) - xy) == 'dot(x, y) + dot(A, y) + -dot(x, y)'
    assert str(xy - (Ay + xy)) == 'dot(x, y) + -dot(A, y) + -dot(x, y)'

    assert str((xy + Ay) - 1) == 'dot(x, y) + dot(A, y) + -1'
    assert str((xy + Ay) + 1) == 'dot(x, y) + dot(A, y) + 1'
    assert str(1 - (xy + Ay)) == '-dot(x, y) + -dot(A, y) + 1'
    assert str(1 + (xy + Ay)) == 'dot(x, y) + dot(A, y) + 1'

    assert str((xy + Ay) + A) == 'dot(x, y) + dot(A, y) + A'


def test_combined_source():
    assert str(x * y) == '((x) * (y)) * 1'
    assert str(x * 2 * y) == '((2 * x) * (y)) * 1'
    assert str((x * y) * 2) == '((x) * (y)) * 2'
    assert str((x * y) * A) == '((x) * (y)) * A'
    assert str((-x * A) * y) == '(((-1 * A) * x) * (y)) * 1'
    assert str(x * y + x) == '((x) * (y)) * 1 + x'
    assert str(x * y + y * x) == '((x) * (y)) * 1 + ((y) * (x)) * 1'


def test_source_addition():
    assert str(x + y) == 'x + y'
    assert str(x + A) == 'x + A'
    assert str(A + x) == 'x + A'
    assert str(x + 1) == 'x + 1'
    assert str(1 + x) == 'x + 1'

    with pytest.raises(TypeError):
        x + []
    with pytest.raises(TypeError):
        x + "1"

    assert str(x - y) == 'x + -1 * y'
    assert str(x - A) == 'x + -A'
    assert str(A - x) == '-1 * x + A'
    assert str(1 - x) == '-1 * x + 1'
    assert str(x - 1) == 'x + -1'

    with pytest.raises(TypeError):
        x - {}
    with pytest.raises(TypeError):
        "1" - x


def test_vector_list():
    xy = x + y

    assert str(xy * 2) == '2 * x + 2 * y'
    assert str((xy * 2) * 2) == '(2 * 2) * x + (2 * 2) * y'
    assert str((xy * 2) / 0.5) == '(2 * 2) * x + (2 * 2) * y'

    assert str(2 * xy) == '2 * x + 2 * y'
    assert str(2 * (xy * 2)) == '(2 * 2) * x + (2 * 2) * y'
    assert str(2 * (2 * xy)) == '(2 * 2) * x + (2 * 2) * y'

    assert str(xy + x) == 'x + y + x'
    assert str(xy + 0.5) == 'x + y + 0.5'
    assert str(xy + A) == 'x + y + A'
    assert str(xy + xy) == 'x + y + x + y'
    assert str(xy + 2 * xy) == 'x + y + 2 * x + 2 * y'

    with pytest.raises(TypeError):
        xy + ''

    assert str(x + xy) == 'x + x + y'
    assert str(0.5 + xy) == 'x + y + 0.5'
    assert str(A + xy) == 'x + y + A'

    with pytest.raises(TypeError):
        '' + xy

    assert str(xy - x) == 'x + y + -1 * x'
    assert str(xy - 0.5) == 'x + y + -0.5'
    assert str(xy - A) == 'x + y + -A'
    assert str(xy - xy) == 'x + y + -1 * x + -1 * y'
    assert str(xy - 2 * xy) == 'x + y + -2 * x + -2 * y'

    with pytest.raises(TypeError):
        xy - ''

    assert str(x - xy) == 'x + -1 * x + -1 * y'
    assert str(0.5 - xy) == '-1 * x + -1 * y + 0.5'
    assert str(A - xy) == '-1 * x + -1 * y + A'

    with pytest.raises(TypeError):
        '' - xy

    assert str(-xy) == '-1 * x + -1 * y'
    assert str(-(-xy)) == 'x + y'



"""
  Licensing:
    This code is distributed under the GNU LGPL license.

  Authors:
    Original FORTRAN77 version of i4_sobol by Bennett Fox.
    MATLAB version by John Burkardt.
    PYTHON version by Corrado Chisari

    Original Python version of is_prime by Corrado Chisari

    Original MATLAB versions of other functions by John Burkardt.
    PYTHON versions by Corrado Chisari

    Original code is available from http://people.sc.fsu.edu/~jburkardt/py_src/sobol/sobol.html
"""

from __future__ import absolute_import

import numpy as np

__all__ = ['i4_bit_hi1', 'i4_bit_lo0', 'i4_sobol_generate',
           'i4_sobol', 'i4_uniform', 'prime_ge', 'is_prime']


def i4_bit_hi1(n):
    """
    i4_bit_hi1 returns the position of the high 1 bit base 2 in an integer.

    Example:
      +------+-------------+-----
      |    N |      Binary | BIT
      +------|-------------+-----
      |    0 |           0 |   0
      |    1 |           1 |   1
      |    2 |          10 |   2
      |    3 |          11 |   2
      |    4 |         100 |   3
      |    5 |         101 |   3
      |    6 |         110 |   3
      |    7 |         111 |   3
      |    8 |        1000 |   4
      |    9 |        1001 |   4
      |   10 |        1010 |   4
      |   11 |        1011 |   4
      |   12 |        1100 |   4
      |   13 |        1101 |   4
      |   14 |        1110 |   4
      |   15 |        1111 |   4
      |   16 |       10000 |   5
      |   17 |       10001 |   5
      | 1023 |  1111111111 |  10
      | 1024 | 10000000000 |  11
      | 1025 | 10000000001 |  11

    Parameters:
      Input, integer N, the integer to be measured.
      N should be nonnegative.  If N is nonpositive, the value will always be 0.

      Output, integer BIT, the number of bits base 2.
    """
    i = np.floor(n)
    bit = 0
    while (1):
        if (i <= 0):
            break
        bit += 1
        i = np.floor(i / 2.)
    return bit


def i4_bit_lo0(n):
    """
    I4_BIT_LO0 returns the position of the low 0 bit base 2 in an integer.

    Example:
      +------+------------+----
      |    N |     Binary | BIT
      +------+------------+----
      |    0 |          0 |   1
      |    1 |          1 |   2
      |    2 |         10 |   1
      |    3 |         11 |   3
      |    4 |        100 |   1
      |    5 |        101 |   2
      |    6 |        110 |   1
      |    7 |        111 |   4
      |    8 |       1000 |   1
      |    9 |       1001 |   2
      |   10 |       1010 |   1
      |   11 |       1011 |   3
      |   12 |       1100 |   1
      |   13 |       1101 |   2
      |   14 |       1110 |   1
      |   15 |       1111 |   5
      |   16 |      10000 |   1
      |   17 |      10001 |   2
      | 1023 | 1111111111 |   1
      | 1024 | 0000000000 |   1
      | 1025 | 0000000001 |   1

    Parameters:
      Input, integer N, the integer to be measured.
      N should be nonnegative.

      Output, integer BIT, the position of the low 1 bit.
    """
    bit = 0
    i = np.floor(n)
    while (1):
        bit = bit + 1
        i2 = np.floor(i / 2.)
        if (i == 2 * i2):
            break

        i = i2
    return bit


def i4_sobol_generate(dim_num, n, skip=1):
    """
    i4_sobol_generate generates a Sobol dataset.

    Parameters:
      Input, integer dim_num, the spatial dimension.
      Input, integer N, the number of points to generate.
      Input, integer SKIP, the number of initial points to skip.

      Output, real R(M,N), the points.
    """
    r = np.full((n, dim_num), np.nan)
    for j in range(n):
        seed = j + 1
        r[j, 0:dim_num], next_seed = i4_sobol(dim_num, seed)

    return r


def i4_sobol(dim_num, seed):
    """
    i4_sobol generates a new quasirandom Sobol vector with each call.

    Discussion:
      The routine adapts the ideas of Antonov and Saleev.

    Reference:
      Antonov, Saleev,
      USSR Computational Mathematics and Mathematical Physics,
      Volume 19, 1980, pages 252 - 256.

      Paul Bratley, Bennett Fox,
      Algorithm 659:
      Implementing Sobol's Quasirandom Sequence Generator,
      ACM Transactions on Mathematical Software,
      Volume 14, Number 1, pages 88-100, 1988.

      Bennett Fox,
      Algorithm 647:
      Implementation and Relative Efficiency of Quasirandom
      Sequence Generators,
      ACM Transactions on Mathematical Software,
      Volume 12, Number 4, pages 362-376, 1986.

      Ilya Sobol,
      USSR Computational Mathematics and Mathematical Physics,
      Volume 16, pages 236-242, 1977.

      Ilya Sobol, Levitan,
      The Production of Points Uniformly Distributed in a Multidimensional
      Cube (in Russian),
      Preprint IPM Akad. Nauk SSSR,
      Number 40, Moscow 1976.

    Parameters:
      Input, integer DIM_NUM, the number of spatial dimensions.
      DIM_NUM must satisfy 1 <= DIM_NUM <= 40.

      Input/output, integer SEED, the "seed" for the sequence.
      This is essentially the index in the sequence of the quasirandom
      value to be generated.  On output, SEED has been set to the
      appropriate next value, usually simply SEED+1.
      If SEED is less than 0 on input, it is treated as though it were 0.
      An input value of 0 requests the first (0-th) element of the sequence.

      Output, real QUASI(DIM_NUM), the next quasirandom vector.
    """
    global atmost
    global dim_max
    global dim_num_save
    global initialized
    global lastq
    global log_max
    global maxcol
    global poly
    global recipd
    global seed_save
    global v

    if 'initialized' not in list(globals().keys()):
        initialized = 0
        dim_num_save = -1

    if (not initialized or dim_num != dim_num_save):
        initialized = 1
        dim_max = 40
        dim_num_save = -1
        log_max = 30
        seed_save = -1

#  Initialize (part of) V.
        v = np.zeros((dim_max, log_max))
        v[0:40, 0] = np.transpose([
            1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
            1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
            1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
            1, 1, 1, 1, 1, 1, 1, 1, 1, 1])

        v[2:40, 1] = np.transpose([
                  1, 3, 1, 3, 1, 3, 3, 1,
            3, 1, 3, 1, 3, 1, 1, 3, 1, 3,
            1, 3, 1, 3, 3, 1, 3, 1, 3, 1,
            3, 1, 1, 3, 1, 3, 1, 3, 1, 3])

        v[3:40, 2] = np.transpose([
                     7, 5, 1, 3, 3, 7, 5,
            5, 7, 7, 1, 3, 3, 7, 5, 1, 1,
            5, 3, 3, 1, 7, 5, 1, 3, 3, 7,
            5, 1, 1, 5, 7, 7, 5, 1, 3, 3])

        v[5:40, 3] = np.transpose([
                                1, 7,  9,  13, 11,
            1, 3,  7,  9,  5,  13, 13, 11, 3,  15,
            5, 3,  15, 7,  9,  13, 9,  1,  11, 7,
            5, 15, 1,  15, 11, 5,  3,  1,  7,  9])

        v[7:40, 4] = np.transpose([
                                        9,  3,  27,
            15, 29, 21, 23, 19, 11, 25, 7,  13, 17,
            1,  25, 29, 3,  31, 11, 5,  23, 27, 19,
            21, 5,  1,  17, 13, 7,  15, 9,  31, 9])

        v[13:40, 5] = np.transpose([
                        37, 33, 7,  5,  11, 39, 63,
            27, 17, 15, 23, 29, 3,  21, 13, 31, 25,
            9,  49, 33, 19, 29, 11, 19, 27, 15, 25])

        v[19:40, 6] = np.transpose([
                                                   13,
            33, 115, 41, 79, 17, 29,  119, 75, 73, 105,
            7,  59,  65, 21, 3,  113, 61,  89, 45, 107])

        v[37:40, 7] = np.transpose([
            7, 23, 39])

#  Set POLY.
        poly = [
            1,   3,   7,   11,  13,  19,  25,  37,  59,  47,
            61,  55,  41,  67,  97,  91,  109, 103, 115, 131,
            193, 137, 145, 143, 241, 157, 185, 167, 229, 171,
            213, 191, 253, 203, 211, 239, 247, 285, 369, 299]

        atmost = 2 ** log_max - 1

#  Find the number of bits in ATMOST.
        maxcol = i4_bit_hi1(atmost)

#  Initialize row 1 of V.
        v[0, 0:maxcol] = 1


#  Things to do only if the dimension changed.
    if (dim_num != dim_num_save):

#  Check parameters.
        if (dim_num < 1 or dim_max < dim_num):
            print('I4_SOBOL - Fatal error!')
            print('  The spatial dimension DIM_NUM should satisfy:')
            print('    1 <= DIM_NUM <= %d' % dim_max)
            print('  But this input value is DIM_NUM = %d' % dim_num)
            return

        dim_num_save = dim_num

#  Initialize the remaining rows of V.
        for i in range(2, dim_num + 1):

#  The bits of the integer POLY(I) gives the form of polynomial I.
#  Find the degree of polynomial I from binary encoding.
            j = poly[i - 1]
            m = 0
            while (1):
                j = np.floor(j / 2.)
                if (j <= 0):
                    break
                m = m + 1

#  Expand this bit pattern to separate components of the logical array INCLUD.
            j = poly[i - 1]
            includ = np.zeros(m)
            for k in range(m, 0, -1):
                j2 = np.floor(j / 2.)
                includ[k - 1] = (j != 2 * j2)
                j = j2

#  Calculate the remaining elements of row I as explained
#  in Bratley and Fox, section 2.
            for j in range(m + 1, maxcol + 1):
                newv = v[i - 1, j - m - 1]
                l = 1
                for k in range(1, m + 1):
                    l = 2 * l
                    if (includ[k - 1]):
                        newv = np.bitwise_xor(
                            int(newv), int(l * v[i - 1, j - k - 1]))
                v[i - 1, j - 1] = newv

#  Multiply columns of V by appropriate power of 2.
        l = 1
        for j in range(maxcol - 1, 0, -1):
            l = 2 * l
            v[0:dim_num, j - 1] = v[0:dim_num, j - 1] * l

#  RECIPD is 1/(common denominator of the elements in V).
        recipd = 1.0 / (2 * l)
        lastq = np.zeros(dim_num)

    seed = int(np.floor(seed))

    if (seed < 0):
        seed = 0

    if (seed == 0):
        l = 1
        lastq = np.zeros(dim_num)

    elif (seed == seed_save + 1):

#  Find the position of the right-hand zero in SEED.
        l = i4_bit_lo0(seed)

    elif (seed <= seed_save):

        seed_save = 0
        l = 1
        lastq = np.zeros(dim_num)

        for seed_temp in range(int(seed_save), int(seed)):
            l = i4_bit_lo0(seed_temp)
            for i in range(1, dim_num + 1):
                lastq[i - 1] = np.bitwise_xor(
                    int(lastq[i - 1]), int(v[i - 1, l - 1]))

        l = i4_bit_lo0(seed)

    elif (seed_save + 1 < seed):

        for seed_temp in range(int(seed_save + 1), int(seed)):
            l = i4_bit_lo0(seed_temp)
            for i in range(1, dim_num + 1):
                lastq[i - 1] = np.bitwise_xor(
                    int(lastq[i - 1]), int(v[i - 1, l - 1]))

        l = i4_bit_lo0(seed)

#  Check that the user is not calling too many times!
    if (maxcol < l):
        print('I4_SOBOL - Fatal error!')
        print('  Too many calls!')
        print('  MAXCOL = %d\n' % maxcol)
        print('  L =      %d\n' % l)
        return

#  Calculate the new components of QUASI.
    quasi = np.zeros(dim_num)
    for i in range(1, dim_num + 1):
        quasi[i - 1] = lastq[i - 1] * recipd
        lastq[i - 1] = np.bitwise_xor(
            int(lastq[i - 1]), int(v[i - 1, l - 1]))

    seed_save = seed
    seed = seed + 1

    return [quasi, seed]


def i4_uniform(a, b, seed):
    """
    i4_uniform returns a scaled pseudorandom I4.

    Discussion:
      The pseudorandom number will be scaled to be uniformly distributed
      between A and B.

    Reference:
      Paul Bratley, Bennett Fox, Linus Schrage,
      A Guide to Simulation,
      Springer Verlag, pages 201-202, 1983.

      Pierre L'Ecuyer,
      Random Number Generation,
      in Handbook of Simulation,
      edited by Jerry Banks,
      Wiley Interscience, page 95, 1998.

      Bennett Fox,
      Algorithm 647:
      Implementation and Relative Efficiency of Quasirandom
      Sequence Generators,
      ACM Transactions on Mathematical Software,
      Volume 12, Number 4, pages 362-376, 1986.

      Peter Lewis, Allen Goodman, James Miller
      A Pseudo-Random Number Generator for the System/360,
      IBM Systems Journal,
      Volume 8, pages 136-143, 1969.

    Parameters:
      Input, integer A, B, the minimum and maximum acceptable values.
      Input, integer SEED, a seed for the random number generator.

      Output, integer C, the randomly chosen integer.
      Output, integer SEED, the updated seed.
    """
    if (seed == 0):
        print('I4_UNIFORM - Fatal error!')
        print('  Input SEED = 0!')

    seed = np.floor(seed)
    a = round(a)
    b = round(b)

    seed = np.mod(seed, 2147483647)

    if (seed < 0):
        seed = seed + 2147483647

    k = np.floor(seed / 127773)

    seed = 16807 * (seed - k * 127773) - k * 2836

    if (seed < 0):
        seed = seed + 2147483647

    r = seed * 4.656612875E-10

#  Scale R to lie between A-0.5 and B+0.5.
    r = (1.0 - r) * (min(a, b) - 0.5) + r * (max(a, b) + 0.5)

#  Use rounding to convert R to an integer between A and B.
    value = round(r)

    value = max(value, min(a, b))
    value = min(value, max(a, b))

    c = value

    return [int(c), int(seed)]


def prime_ge(n):
    """
    PRIME_GE returns the smallest prime greater than or equal to N.

    Example:
      +-----+---------
      |   N | PRIME_GE
      +-----+---------
      | -10 |        2
      |   1 |        2
      |   2 |        2
      |   3 |        3
      |   4 |        5
      |   5 |        5
      |   6 |        7
      |   7 |        7
      |   8 |       11
      |   9 |       11
      |  10 |       11

    Parameters:
      Input, integer N, the number to be bounded.

      Output, integer P, the smallest prime number that is greater
      than or equal to N.
    """
    p = max(np.ceil(n), 2)
    while (not is_prime(p)):
        p = p + 1

    return p


def is_prime(n):
    """
    is_prime returns True if N is a prime number, False otherwise

    Parameters:
       Input, integer N, the number to be checked.

       Output, boolean value, True or False
    """
    if n != int(n) or n < 1:
        return False
    p = 2
    while p < n:
        if n % p == 0:
            return False
        p += 1
    return True



import nengo
from .compat import iteritems


def probe_all(net, recursive=False, probe_options=None,  # noqa: C901
              **probe_args):
    """Probes all objects in a network.

    Parameters
    ----------
    net : nengo.Network
    recursive : bool, optional
        probe subnetworks recursively, False by default.
    probe_options: dict, optional
        A dict of the form {nengo_object_class: [attributes_to_probe]}.
        If not specified, every probeable attribute of every object
        will be probed.

    Returns
    -------
    A dictionary that maps objects and their attributes to their probes.

    Examples
    --------

    Probe the decoded output and spikes in all ensembles in a network and
    its subnetworks.

    ::

        model = nengo.Network(label='_test_probing')
        with model:
            ens1 = nengo.Ensemble(n_neurons=1, dimensions=1)
            node1 = nengo.Node(output=[0])
            conn = nengo.Connection(node1, ens1)
            subnet = nengo.Network(label='subnet')
            with subnet:
                ens2 = nengo.Ensemble(n_neurons=1, dimensions=1)
                node2 = nengo.Node(output=[0])

        probe_options = {nengo.Ensemble: ['decoded_output', 'spikes']}
        object_probe_dict = probe(model, recursive=True,
                                  probe_options=probe_options)

    """

    probes = {}

    def probe_helper(net, recursive, probe_options):
        with net:
            for obj_type, obj_list in iteritems(net.objects):

                # recursively probe subnetworks if required
                if obj_type is nengo.Network and recursive:
                    for subnet in obj_list:
                        probe_helper(subnet, recursive=recursive,
                                     probe_options=probe_options)

                # probe all probeable objects
                elif probe_options is None:
                    for obj in obj_list:
                        if hasattr(obj, 'probeable') and len(
                                obj.probeable) > 0:
                            probes[obj] = {}
                            for probeable in obj.probeable:
                                probes[obj][probeable] = nengo.Probe(
                                    obj, probeable, **probe_args)

                # probe specified objects only
                elif obj_type in probe_options:
                    for obj in obj_list:
                        probes[obj] = {}
                        for attr in probe_options[obj_type]:
                            probes[obj][attr] = (
                                nengo.Probe(obj, attr, **probe_args))

    probe_helper(net, recursive, probe_options)
    return probes



from __future__ import division, absolute_import, print_function

from os.path import join, sep, dirname

from numpy.distutils.misc_util import (
    appendpath, minrelpath, gpaths, get_shared_lib_extension, get_info
    )
from numpy.testing import (
    assert_, assert_equal
    )

ajoin = lambda *paths: join(*((sep,)+paths))

class TestAppendpath(object):

    def test_1(self):
        assert_equal(appendpath('prefix', 'name'), join('prefix', 'name'))
        assert_equal(appendpath('/prefix', 'name'), ajoin('prefix', 'name'))
        assert_equal(appendpath('/prefix', '/name'), ajoin('prefix', 'name'))
        assert_equal(appendpath('prefix', '/name'), join('prefix', 'name'))

    def test_2(self):
        assert_equal(appendpath('prefix/sub', 'name'),
                     join('prefix', 'sub', 'name'))
        assert_equal(appendpath('prefix/sub', 'sup/name'),
                     join('prefix', 'sub', 'sup', 'name'))
        assert_equal(appendpath('/prefix/sub', '/prefix/name'),
                     ajoin('prefix', 'sub', 'name'))

    def test_3(self):
        assert_equal(appendpath('/prefix/sub', '/prefix/sup/name'),
                     ajoin('prefix', 'sub', 'sup', 'name'))
        assert_equal(appendpath('/prefix/sub/sub2', '/prefix/sup/sup2/name'),
                     ajoin('prefix', 'sub', 'sub2', 'sup', 'sup2', 'name'))
        assert_equal(appendpath('/prefix/sub/sub2', '/prefix/sub/sup/name'),
                     ajoin('prefix', 'sub', 'sub2', 'sup', 'name'))

class TestMinrelpath(object):

    def test_1(self):
        n = lambda path: path.replace('/', sep)
        assert_equal(minrelpath(n('aa/bb')), n('aa/bb'))
        assert_equal(minrelpath('..'), '..')
        assert_equal(minrelpath(n('aa/..')), '')
        assert_equal(minrelpath(n('aa/../bb')), 'bb')
        assert_equal(minrelpath(n('aa/bb/..')), 'aa')
        assert_equal(minrelpath(n('aa/bb/../..')), '')
        assert_equal(minrelpath(n('aa/bb/../cc/../dd')), n('aa/dd'))
        assert_equal(minrelpath(n('.././..')), n('../..'))
        assert_equal(minrelpath(n('aa/bb/.././../dd')), n('dd'))

class TestGpaths(object):

    def test_gpaths(self):
        local_path = minrelpath(join(dirname(__file__), '..'))
        ls = gpaths('command/*.py', local_path)
        assert_(join(local_path, 'command', 'build_src.py') in ls, repr(ls))
        f = gpaths('system_info.py', local_path)
        assert_(join(local_path, 'system_info.py') == f[0], repr(f))

class TestSharedExtension(object):

    def test_get_shared_lib_extension(self):
        import sys
        ext = get_shared_lib_extension(is_python_ext=False)
        if sys.platform.startswith('linux'):
            assert_equal(ext, '.so')
        elif sys.platform.startswith('gnukfreebsd'):
            assert_equal(ext, '.so')
        elif sys.platform.startswith('darwin'):
            assert_equal(ext, '.dylib')
        elif sys.platform.startswith('win'):
            assert_equal(ext, '.dll')
        # just check for no crash
        assert_(get_shared_lib_extension(is_python_ext=True))


def test_installed_npymath_ini():
    # Regression test for gh-7707.  If npymath.ini wasn't installed, then this
    # will give an error.
    info = get_info('npymath')

    assert isinstance(info, dict)
    assert "define_macros" in info



import warnings

import nengo
from nengo.spa.buffer import Buffer


class Memory(Buffer):
    """A SPA module capable of storing a vector over time.

    Parameters are the same as `.spa.Buffer`, with the addition of
    ``synapse`` and ``tau``.

    .. note:: Deprecated in Nengo 2.1.0. Use `.spa.State` instead.

    Parameters
    ----------
    dimensions : int
        Number of dimensions for the vector.
    subdimensions : int, optional (Default: 16)
        Size of the individual ensembles making up the vector.
        Must divide ``dimensions`` evenly.
    neurons_per_dimensions : int, optional (Default: 50)
        Number of neurons in an ensemble will be
        ``neurons_per_dimensions * subdimensions``.
    synapse : float, optional (Default: 0.01)
        Synaptic filter to use on recurrent connection.
    vocab : Vocabulary, optional (Default: None)
        The vocabulary to use to interpret the vector. If None,
        the default vocabulary for the given dimensionality is used.
    tau : float or None, optional (Default: None)
        Effective time constant of the integrator. If None, it should
        have an infinite time constant.
    direct : bool, optional (Default: False)
        Whether or not to use direct mode for the neurons.

    label : str, optional (Default: None)
        A name for the ensemble. Used for debugging and visualization.
    seed : int, optional (Default: None)
        The seed used for random number generation.
    add_to_container : bool, optional (Default: None)
        Determines if this Network will be added to the current container.
        If None, will be true if currently within a Network.
    """

    def __init__(self, dimensions, subdimensions=16, neurons_per_dimension=50,
                 synapse=0.01, vocab=None, tau=None, direct=False,
                 label=None, seed=None, add_to_container=None):
        warnings.warn("Memory is deprecated in favour of spa.State",
                      DeprecationWarning)
        super(Memory, self).__init__(
            dimensions=dimensions,
            subdimensions=subdimensions,
            neurons_per_dimension=neurons_per_dimension,
            vocab=vocab,
            direct=direct,
            label=label,
            seed=seed,
            add_to_container=add_to_container)

        if tau is None:
            transform = 1.0
        else:
            transform = 1.0 - synapse / tau

        with self:
            nengo.Connection(self.state.output, self.state.input,
                             transform=transform, synapse=synapse)
